"""Generic driver for reading and writing DataFrames."""

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, ClassVar

import pandas as pd

from hydromt.data_catalog.drivers import BaseDriver
from hydromt.error import NoDataStrategy
from hydromt.typing import (
    Variables,
)

logger = logging.getLogger(__name__)


class DataFrameDriver(BaseDriver, ABC):
    """Abstract Driver to read DataFrames."""

    supports_writing: ClassVar[bool] = False

    @abstractmethod
    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        variables: Variables | None = None,
    ) -> pd.DataFrame:
        """
        Read data from one or more URIs into a pandas DataFrame.

        This abstract method defines the interface for all DataFrame-based drivers.
        Subclasses should implement data loading logic appropriate for the format being read.

        Parameters
        ----------
        uris : list[str]
            List of URIs to read data from. The driver decides how to handle multiple files.
        handle_nodata : NoDataStrategy, optional
            Strategy to handle missing or empty data. Default is NoDataStrategy.RAISE.
        variables : Variables | None, optional
            List of variable names (columns) to select from the source data. Default is None.

        Returns
        -------
        pd.DataFrame
            The loaded DataFrame.

        """
        ...

    @abstractmethod
    def write(
        self,
        path: Path | str,
        data: pd.DataFrame,
        *,
        write_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write a pandas DataFrame to a file.

        This abstract method defines the interface for writing DataFrames in all DataFrame-based drivers.
        Subclasses should implement logic appropriate for the target file format (e.g., CSV, Parquet, Excel).

        Parameters
        ----------
        path : Path | str
            Destination path or URI where the DataFrame should be written.
        data : pd.DataFrame
            The DataFrame to be written to disk.
        write_kwargs : dict[str, Any], optional
            Additional keyword arguments to pass to the underlying pandas write function
            (e.g., `to_csv`, `to_excel`, `to_parquet`). Default is None.

        Returns
        -------
        Path
            The path where the data was written.

        """
        ...



"""DataSource class for the RasterDataset type."""

import logging
from os.path import basename, splitext
from pathlib import Path
from typing import Any, ClassVar, List, Literal, Optional

import numpy as np
import xarray as xr
from pydantic import Field
from pyproj.exceptions import CRSError
from pystac import Asset as StacAsset
from pystac import Catalog as StacCatalog
from pystac import Item as StacItem
from pystac import MediaType

from hydromt.data_catalog.adapters.rasterdataset import RasterDatasetAdapter
from hydromt.data_catalog.drivers import RasterDatasetDriver
from hydromt.data_catalog.sources.data_source import DataSource
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.gis_utils import _parse_geom_bbox_buffer
from hydromt.typing import (
    Bbox,
    Geom,
    TimeRange,
    TotalBounds,
    Zoom,
)
from hydromt.typing.fsspec_types import FSSpecFileSystem

logger = logging.getLogger(__name__)


class RasterDatasetSource(DataSource):
    """DataSource class for the RasterDataset type."""

    data_type: ClassVar[Literal["RasterDataset"]] = "RasterDataset"
    _fallback_driver_read: ClassVar[str] = "rasterio"
    _fallback_driver_write: ClassVar[str] = "raster_xarray"
    driver: RasterDatasetDriver
    data_adapter: RasterDatasetAdapter = Field(default_factory=RasterDatasetAdapter)

    def read_data(
        self,
        *,
        bbox: Optional[Bbox] = None,
        mask: Optional[Geom] = None,
        buffer: int = 0,
        variables: Optional[List[str]] = None,
        time_range: Optional[TimeRange] = None,
        zoom: Optional[Zoom] = None,
        chunks: Optional[dict] = None,
        single_var_as_array: bool = True,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> xr.Dataset | xr.DataArray | None:
        """
        Read data from this source.

        Args:
        """
        self._mark_as_used()
        self._log_start_read_data()

        if bbox is not None:
            # buffer will be applied in transform
            mask = _parse_geom_bbox_buffer(mask, bbox)

        # Transform time_range and variables to match the data source
        src_time_range = self.data_adapter._to_source_timerange(time_range)
        vrs = self.data_adapter._to_source_variables(variables)

        uris: List[str] = self.uri_resolver.resolve(
            self.full_uri,
            time_range=src_time_range,
            mask=mask,
            variables=vrs,
            zoom=zoom,
            metadata=self.metadata,
            handle_nodata=handle_nodata,
        )
        if not uris:
            return None  # handle_nodata == ignore

        ds: xr.Dataset = self.driver.read(
            uris,
            handle_nodata=handle_nodata,
            mask=mask,
            variables=vrs,
            zoom=zoom,
            chunks=chunks,
            metadata=self.metadata,
        )
        if ds is None:
            return None  # handle_nodata == ignore

        return self.data_adapter.transform(
            ds,
            self.metadata,
            mask=mask,
            variables=variables,
            time_range=time_range,
            single_var_as_array=single_var_as_array,
            buffer=buffer,
            handle_nodata=handle_nodata,
        )

    def to_file(
        self,
        file_path: Path | str,
        *,
        driver_override: RasterDatasetDriver | None = None,
        bbox: Bbox | None = None,
        mask: Geom | None = None,
        buffer: int = 0,
        time_range: TimeRange | None = None,
        zoom: Zoom | None = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        write_kwargs: dict[str, Any] | None = None,
        variables: List[str] | None = None,
    ) -> "RasterDatasetSource | None":
        """
        Write the RasterDatasetSource to a local file.

        args:
        """
        if driver_override is None and not self.driver.supports_writing:
            # default to fallback driver
            driver = RasterDatasetDriver.model_validate(self._fallback_driver_write)
        elif driver_override:
            if not driver_override.supports_writing:
                raise RuntimeError(
                    f"driver: '{driver_override.name}' does not support writing data."
                )
            driver = driver_override
        else:
            # use local filesystem
            driver = self.driver.model_copy(
                update={"filesystem": FSSpecFileSystem.create("local")}
            )

        ds = self.read_data(
            bbox=bbox,
            mask=mask,
            buffer=buffer,
            time_range=time_range,
            zoom=zoom,
            handle_nodata=handle_nodata,
            variables=variables,
        )
        if ds is None:  # handle_nodata == ignore
            return None

        file_path = Path(file_path)

        dest_path = driver.write(
            file_path,
            ds,
            write_kwargs=write_kwargs,
        )

        # update driver based on local path
        update = {
            "uri": dest_path.as_posix(),
            "root": None,
            "driver": driver,
        }

        return self.model_copy(update=update)

    def _detect_bbox(
        self,
        *,
        strict: bool = False,
        ds: Optional[xr.Dataset] = None,
    ) -> TotalBounds:
        """Detect the bounding box and crs of the dataset.

        If no dataset is provided, it will be fetched according to the settings in the
        adapter. also see :py:meth:`hydromt.RasterdatasetAdapter.get_data`. the
        coordinates are in the CRS of the dataset itself, which is also returned
        alongside the coordinates.


        Parameters
        ----------
        ds: xr.Dataset, xr.DataArray, Optional
            the dataset to detect the bounding box of.
            If none is provided, :py:meth:`hydromt.RasterdatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        bbox: Tuple[np.float64,np.float64,np.float64,np.float64]
            the bounding box coordinates of the data. coordinates are returned as
            [xmin,ymin,xmax,ymax]
        crs: int
            The ESPG code of the CRS of the coordinates returned in bbox
        """
        if ds is None:
            ds = self.read_data()
        crs = ds.raster.crs.to_epsg()
        bounds = ds.raster.bounds

        return bounds, crs

    def _detect_time_range(
        self,
        *,
        strict: bool = False,
        ds: Optional[xr.Dataset] = None,
    ) -> TimeRange:
        """Detect the temporal range of the dataset.

        If no dataset is provided, it will be fetched accodring to the settings in the
        addapter. also see :py:meth:`hydromt.RasterdatasetAdapter.get_data`.

        Parameters
        ----------
        ds: xr.Dataset, xr.DataArray, Optional
            the dataset to detect the time range of. It must have a time dimentsion set.
            If none is provided, :py:meth:`hydromt.RasterdatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        range: TimeRange
            A tuple containing the start and end of the time dimension. Range is
            inclusive on both sides.
        """
        if ds is None:
            ds = self.read_data()

        if not np.issubdtype(ds[ds.raster.time_dim].dtype, np.datetime64):
            ds = ds.convert_calendar("standard")

        return TimeRange(
            start=ds[ds.raster.time_dim].values.min(),
            end=ds[ds.raster.time_dim].values.max(),
        )

    def to_stac_catalog(
        self,
        handle_nodata: NoDataStrategy = NoDataStrategy.WARN,
    ) -> Optional[StacCatalog]:
        """
        Convert a rasterdataset into a STAC Catalog representation.

        The collection will contain an asset for each of the associated files.


        Parameters
        ----------
        - handle_nodata (str, optional): The error handling strategy.
          Options are: "raise" to raise an error on failure, "warn" to log a warning, "ignore" to skip the
          dataset on failure

        Returns
        -------
        - Optional[StacCatalog]: The STAC Catalog representation of the dataset, or None
          if the dataset was skipped.
        """
        try:
            bbox, crs = self.get_bbox(detect=True)
            bbox = list(bbox)
            time_range = self.get_time_range(detect=True)
            props = {**self.metadata.model_dump(exclude_none=True), "crs": crs}
            ext = splitext(self.full_uri)[-1]
            if ext == ".nc" or ext == ".vrt":
                media_type = MediaType.HDF5
            elif ext == ".tiff":
                media_type = MediaType.TIFF
            elif ext == ".cog":
                media_type = MediaType.COG
            elif ext == ".png":
                media_type = MediaType.PNG
            else:
                raise RuntimeError(
                    f"Unknown extension: {ext} cannot determine media type"
                )
        except (IndexError, KeyError, CRSError):
            exec_nodata_strat(
                f"Skipping {self.name} during stac conversion because detecting spacial extent failed.",
                strategy=handle_nodata,
            )
            return None

        else:
            # else makes type checkers a bit happier
            stac_catalog = StacCatalog(
                self.name,
                description=self.name,
            )
            stac_item = StacItem(
                self.name,
                geometry=None,
                bbox=list(bbox),
                properties=props,
                datetime=None,
                start_datetime=time_range.start,
                end_datetime=time_range.end,
            )
            stac_asset = StacAsset(str(self.full_uri), media_type=media_type)
            base_name = basename(self.full_uri)
            stac_item.add_asset(base_name, stac_asset)

            stac_catalog.add_item(stac_item)
            return stac_catalog

    @classmethod
    def _infer_default_driver(
        cls,
        uri: str | None = None,
    ) -> str:
        return super()._infer_default_driver(uri, RasterDatasetDriver)



def _partition_dictionaries(left, right):
    """Calculate a partitioning of the two dictionaries.

    given dictionaries A and B this function will the following partition:
    (A ∩ B, A - B, B - A)
    """
    common = {}
    left_less_right = {}
    right_less_left = {}
    key_union = set(left.keys()) | set(right.keys())

    for key in key_union:
        value_left = left.get(key, None)
        value_right = right.get(key, None)
        if isinstance(value_left, dict) and isinstance(value_right, dict):
            (
                common_children,
                unique_left_children,
                unique_right_children,
            ) = _partition_dictionaries(value_left, value_right)
            common[key] = common_children
            if unique_left_children != unique_right_children:
                left_less_right[key] = unique_left_children
                right_less_left[key] = unique_right_children
        elif value_left == value_right:
            common[key] = value_left
        else:
            if value_left is not None:
                left_less_right[key] = value_left
            if value_right is not None:
                right_less_left[key] = value_right

    return common, left_less_right, right_less_left



"""Implementation of the predefined data catalogs entry points."""

import logging
import shutil
import sys
from pathlib import Path
from typing import Callable, ClassVar, Optional

import packaging.version
import pooch

from hydromt._utils.caching import copy_to_local
from hydromt._utils.uris import _is_valid_url
from hydromt.config import SETTINGS

logger = logging.getLogger(__name__)

# get repos folder
_hydromt_root = Path(__file__).parent.parent.parent
if Path(_hydromt_root, "data/catalogs").exists():
    # get local repos catalogs if these exist
    CATALOG_ROOT = Path(_hydromt_root, "data/catalogs").as_posix()
else:
    # otherwise use the online catalogs
    _git_root = r"https://raw.githubusercontent.com/Deltares/hydromt/main"
    CATALOG_ROOT = f"{_git_root}/data/catalogs"

__all__ = [
    "PredefinedCatalog",
    "DeltaresDataCatalog",
    "ArtifactDataCatalog",
    "AWSDataCatalog",
    "GCSCMIP6DataCatalog",
    "create_registry_file",
]

__hydromt_eps__ = [
    "DeltaresDataCatalog",
    "ArtifactDataCatalog",
    "AWSDataCatalog",
    "GCSCMIP6DataCatalog",
]


def create_registry_file(root: Path, registry_path: Optional[Path] = None) -> None:
    """Create a registry file for all catalog files in the root directory.

    The root directory should contain a <version>/data_catalog.yml file per version.
    By default the root directory is the cache directory of the catalog instance.

    Parameters
    ----------
    root: Path
        root directory to search for data_catalog.yml files
    """
    # we don't use pooch.create_registry here as we want to only include vaild data_catalog.yml files
    registry = {}
    for path in root.glob("**/data_catalog.yml"):
        key = path.relative_to(root).as_posix()
        if not _valid_key(key):
            raise ValueError(f"No valid version found in {key}")
        if sys.platform == "win32":
            # The line endings need to be replaced when operating from windows in order to maintain equality of hashes
            _replace_line_endings(path)
        file_hash = pooch.file_hash(path)
        registry[key] = file_hash

    if not registry:
        raise FileNotFoundError(f"No data_catalog.yml files found in {root}")

    if registry_path is None:
        registry_path = Path(root / "registry.txt")
    with open(registry_path, "w") as f:
        for fname, hash in registry.items():
            f.write(f"{fname} {hash}\n")


class PredefinedCatalog(object):
    """Predefined data catalog.

    A predefined data catalog is a collection of data_catalog.yml files that are stored in a
    specific directory structure. The catalog is defined by a base_url and a name. The predefined
    catalog can be used to retrieve data_catalog.yml files for specific versions.

    Directory structure:
    - <base_url>/registry.txt
    - <base_url>/<version>/data_catalog.yml

    Cached directory structure:
    - <cache_dir>/<name>/registry.txt
    - <cache_dir>/<name>/<version>/data_catalog.yml
    """

    # required class variables to be defined in subclasses
    base_url: ClassVar[str] = CATALOG_ROOT
    name: ClassVar[str] = "predefined_catalog"

    def __init__(
        self, format_version: str = "v0", cache_dir=SETTINGS.cache_root
    ) -> None:
        # init arguments passed by DataCatalog
        self._format_version = format_version
        self._cache_dir: Path = Path(cache_dir)
        # placeholders set by the class
        self._pooch: Optional[pooch.Pooch] = None
        self._versions: Optional[list[str]] = None

    @property
    def registry(self) -> dict:
        """Return the registry."""
        return self.pooch.registry

    @property
    def pooch(self) -> pooch.Pooch:
        """Return a pooch instance with all data catalog files in registry."""
        if self._pooch is None:
            self._create_pooch()
            self._load_registry_file()
        return self._pooch

    @property
    def versions(self) -> list[str]:
        """Return the versions of the catalog."""
        if not self._versions:
            self._versions = self._get_versions()
        return self._versions

    def _create_pooch(self) -> None:
        self._pooch = pooch.create(
            path=self._cache_dir / self.name,
            base_url=self.base_url,
            retry_if_failed=3,
        )

    def _get_versions(self) -> list[str]:
        """Set valid catalog versions."""
        # parse versions from registry, assume registry key is <version>/data_catalog.yml
        # keep only versions that match the format_version
        keys = self.registry.keys()
        _versions = [
            v.split("/")[0] for v in keys if _valid_key(v, self._format_version)
        ]
        if len(_versions) == 0:
            raise RuntimeError(
                f"No compatible catalog version could be found for {self.name}."
            )
        self._versions = sorted(_versions, key=packaging.version.parse)
        return self._versions

    def _load_registry_file(self, overwrite: bool = False) -> None:
        """Create a catalog from a yaml file."""
        if self._pooch is None:
            self._create_pooch()
        if self.registry and not overwrite:
            return
        registry_path = Path(self._cache_dir / self.name / "registry.txt")
        if registry_path.exists():
            registry_path.unlink()
        try:  # try to retrieve and cache the registry file
            copy_to_local(f"{self.base_url}/registry.txt", registry_path)
        except (ConnectionError, FileNotFoundError):
            logger.warning(
                f"Failed to retrieve {self.name} versions file from {self.base_url}."
                " Creating registry file from cached catalog files."
            )
            create_registry_file(registry_path.parent)
        if not registry_path.exists():
            raise FileNotFoundError(
                f"No cached file found. Failed to retrieve {self.name} versions file"
            )
        self.pooch.load_registry(registry_path)

    def get_catalog_file(self, version: Optional[str] = None) -> Optional[Path]:
        """Get the cached catalog file path for a specific version.

        Parameters
        ----------
        version: str, optional
            The version of the catalog to retrieve. If None, the latest version is retrieved.

        Returns
        -------
        Path
            The path to the cachd catalog file.
        """
        if version is None or version == "latest":  # get latest version
            version = self.versions[-1]
        if version not in self.versions:
            raise ValueError(f"Version {version} not found in {self.name} catalog")
        # get the catalog file
        key = f"{version}/data_catalog.yml"
        # fetch the file (download if not cached)
        path = self.pooch.fetch(key, downloader=self._downloader)
        return Path(path) if path else None

    @property
    def _downloader(self) -> Optional[Callable]:
        if not _is_valid_url(self.base_url):
            return _copy_file
        return None


def _valid_key(v: str, format_version: Optional[str] = None) -> bool:
    """Check if the key is a valid version."""
    try:
        packaging.version.parse(v.split("/")[0])
        return v.startswith(format_version) if format_version else True
    except (packaging.version.InvalidVersion, AttributeError):
        return False


def _copy_file(
    url: str,
    output_file: str,
    pooch: Optional[pooch.Pooch] = None,
    check_only: bool = False,
):
    """Copy a local file to the cache directory for testing purposes.

    for more info, see: https://www.fatiando.org/pooch/latest/downloaders.html
    """
    url = Path(url)
    output_file = Path(output_file)
    file_exists = url.is_file()
    if check_only:
        return file_exists
    if not file_exists:
        raise FileNotFoundError(f"Local file {url} does not exist.")
    output_file.parent.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(url, output_file)
    return output_file


class DeltaresDataCatalog(PredefinedCatalog):
    """Deltares data catalog."""

    base_url = f"{CATALOG_ROOT}/deltares_data"
    name = "deltares_data"


class ArtifactDataCatalog(PredefinedCatalog):
    """Artifact data catalog."""

    base_url = f"{CATALOG_ROOT}/artifact_data"
    name = "artifact_data"


class AWSDataCatalog(PredefinedCatalog):
    """AWS data catalog."""

    base_url = f"{CATALOG_ROOT}/aws_data"
    name = "aws_data"


class GCSCMIP6DataCatalog(PredefinedCatalog):
    """GCS CMIP6 data catalog."""

    base_url = f"{CATALOG_ROOT}/gcs_cmip6_data"
    name = "gcs_cmip6_data"


def _replace_line_endings(file_path: Path):
    WINDOWS_LINE_ENDING = b"\r\n"
    UNIX_LINE_ENDING = b"\n"
    with open(file_path, "rb") as open_file:
        content = open_file.read()
    content = content.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)
    with open(file_path, "wb") as open_file:
        open_file.write(content)



"""DataSource class for the Dataset type."""

import logging
from os.path import basename, splitext
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Literal, Optional, Union

import xarray as xr
from pydantic import Field
from pyproj.exceptions import CRSError
from pystac import Asset as StacAsset
from pystac import Catalog as StacCatalog
from pystac import Item as StacItem
from pystac import MediaType

from hydromt.data_catalog.adapters.dataset import DatasetAdapter
from hydromt.data_catalog.drivers import DatasetDriver
from hydromt.data_catalog.sources.data_source import DataSource
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import (
    TimeRange,
)
from hydromt.typing.fsspec_types import FSSpecFileSystem

logger = logging.getLogger(__name__)


class DatasetSource(DataSource):
    """DataSource class for the Dataset type."""

    data_type: ClassVar[Literal["Dataset"]] = "Dataset"
    _fallback_driver_read: ClassVar[str] = "dataset_xarray"
    _fallback_driver_write: ClassVar[str] = "dataset_xarray"
    driver: DatasetDriver
    data_adapter: DatasetAdapter = Field(default_factory=DatasetAdapter)
    zoom_levels: Optional[Dict[int, float]] = None

    def read_data(
        self,
        *,
        variables: Optional[List[str]] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        single_var_as_array: bool = True,
    ) -> xr.Dataset | None:
        """Use the resolver, driver, and data adapter to read and harmonize the data.

        Parameters
        ----------
        variables : Optional[List[str]], optional
            Names of variables to return, or all if None, by default None
            variables queried for, by default None
        time_range : Optional[TimeRange], optional
            left-inclusive start end time of the data, by default None
        handle_nodata : NoDataStrategy, optional
            how to react when no data is found, by default NoDataStrategy.RAISE
        single_var_as_array : bool, optional
            If only a single variable is requested, return as DataArray instead of
            Dataset, by default True

        Returns
        -------
        Union[xr.Dataset, xr.DataArray]
            harmonized data
        """
        self._mark_as_used()
        self._log_start_read_data()

        # Transform time_range and variables to match the data source
        src_time_range = self.data_adapter._to_source_timerange(time_range)
        vrs = self.data_adapter._to_source_variables(variables)

        uris: List[str] = self.uri_resolver.resolve(
            self.full_uri,
            time_range=src_time_range,
            variables=vrs,
            handle_nodata=handle_nodata,
        )
        if not uris:
            return None  # handle_nodata == ignore

        ds: xr.Dataset = self.driver.read(uris, handle_nodata=handle_nodata)
        if ds is None:  # handle_nodata == ignore
            return None

        return self.data_adapter.transform(
            ds,
            self.metadata,
            variables=variables,
            time_range=time_range,
            single_var_as_array=single_var_as_array,
            handle_nodata=handle_nodata,
        )

    def to_file(
        self,
        file_path: Path | str,
        *,
        driver_override: DatasetDriver | None = None,
        time_range: TimeRange | None = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        write_kwargs: dict[str, Any] | None = None,
    ) -> "DatasetSource | None":
        """
        Write the DatasetSource to a local file.

        args:
        """
        if driver_override is None and not self.driver.supports_writing:
            # default to fallback driver
            driver = DatasetDriver.model_validate(self._fallback_driver_write)
        elif driver_override:
            if not driver_override.supports_writing:
                raise RuntimeError(
                    f"driver: '{driver_override.name}' does not support writing data."
                )
            driver = driver_override
        else:
            # use local filesystem
            driver = self.driver.model_copy(
                update={"filesystem": FSSpecFileSystem.create("local")}
            )

        ds: Optional[xr.Dataset] = self.read_data(
            time_range=time_range, handle_nodata=handle_nodata
        )
        if ds is None:  # handle_nodata == ignore
            return None

        # driver can return different path if file ext changes
        dest_path = driver.write(
            file_path,
            ds,
            write_kwargs=write_kwargs,
        )

        # update driver based on local path
        update = {
            "uri": dest_path.as_posix(),
            "root": None,
            "driver": driver,
        }

        return self.model_copy(update=update)

    def _detect_time_range(
        self,
        *,
        strict: bool = False,
        ds: Union[xr.DataArray, xr.Dataset] = None,
    ) -> TimeRange | None:
        """Get the temporal range of a dataset.

        Parameters
        ----------
        ds : Optional[xr.DataArray  |  xr.Dataset]
            The dataset to detect the time range of. It must have a time dimension set.
            If none is provided, :py:meth:`hydromt.DatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        range: TimeRange, optional
            Instance containing the start and end of the time dimension. Range is
            inclusive on both sides. None if the dataset has no time dimension.
        """
        if ds is None:
            ds = self.read_data()

        try:
            return TimeRange(
                start=ds.time[0].values,
                end=ds.time[-1].values,
            )
        except AttributeError:
            raise AttributeError("Dataset has no dimension called 'time'")

    def to_stac_catalog(
        self,
        handle_nodata: NoDataStrategy = NoDataStrategy.WARN,
    ) -> Optional[StacCatalog]:
        """
        Convert a dataset into a STAC Catalog representation.

        The collection will contain an asset for each of the associated files.


        Parameters
        ----------
        - handle_nodata (str, optional): The error handling strategy.
          Options are: "raise" to raise an error on failure, "skip" to skip the
          dataset on failure.

        Returns
        -------
        - Optional[StacCatalog]: The STAC Catalog representation of the dataset, or None
          if the dataset was skipped.
        """
        try:
            time_range = self.get_time_range(detect=True)
            props = {**self.metadata.model_dump(exclude_none=True, exclude_unset=True)}
            ext = splitext(self.full_uri)[-1]
            if ext == ".nc":
                media_type = MediaType.HDF5
            elif ext == ".zarr":
                raise ValueError("STAC does not support zarr datasets")
            else:
                raise ValueError(
                    f"Unknown extension: {ext} cannot determine media type"
                )
        except (IndexError, KeyError, CRSError):
            exec_nodata_strat(
                f"Skipping {self.name} during stac conversion because detecting spacial extent failed.",
                handle_nodata,
            )
            return None

        stac_catalog = StacCatalog(
            self.name,
            description=self.name,
        )
        stac_item = StacItem(
            self.name,
            geometry=None,
            bbox=None,
            properties=props,
            datetime=None,
            start_datetime=time_range.start,
            end_datetime=time_range.end,
        )
        stac_asset = StacAsset(str(self.full_uri), media_type=media_type)
        base_name = basename(self.full_uri)
        stac_item.add_asset(base_name, stac_asset)

        stac_catalog.add_item(stac_item)
        return stac_catalog



"""Geoms component."""

import logging
from typing import TYPE_CHECKING, cast

import geopandas as gpd
from geopandas import GeoDataFrame, GeoSeries
from geopandas.testing import assert_geodataframe_equal

from hydromt.model.components.base import ModelComponent
from hydromt.model.components.spatial import SpatialModelComponent
from hydromt.model.steps import hydromt_step
from hydromt.readers import _expand_wildcards_and_name_placeholder
from hydromt.typing.crs import CRS

if TYPE_CHECKING:
    from hydromt.model.model import Model

logger = logging.getLogger(__name__)


class GeomsComponent(SpatialModelComponent):
    """A component to manage geo-spatial geometries.

    It contains a dictionary of geopandas GeoDataFrames.
    """

    def __init__(
        self,
        model: "Model",
        *,
        filename: str = "geoms/{name}.geojson",
        region_component: str | None = None,
        region_filename: str = "geoms/geoms_region.geojson",
    ):
        """Initialize a GeomsComponent.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        filename: str
            The path to use for reading and writing of component data by default.
            by default ``geoms/{name}.geojson`` ie one file per geodataframe in the data dictionary.
        region_component: str, optional
            The name of the region component to use as reference for this component's
            region. If None, the region will be set to the union of all geometries in
            the data dictionary.
        region_filename: str
            The path to use for writing the region data to a file. By default
            ``geoms/geoms_region.geojson``.
        """
        self._data: dict[str, GeoDataFrame | GeoSeries] = None
        self._filename: str = filename
        super().__init__(
            model=model,
            region_component=region_component,
            region_filename=region_filename,
        )

    @property
    def data(self) -> dict[str, GeoDataFrame | GeoSeries]:
        """Model geometries.

        Return dict of geopandas.GeoDataFrame or geopandas.GeoSeries
        """
        if self._data is None:
            self._initialize()

        assert self._data is not None
        return self._data

    def _initialize(self, skip_read=False) -> None:
        """Initialize geoms."""
        if self._data is None:
            self._data = dict()
            if self.root.is_reading_mode() and not skip_read:
                self.read()

    @property
    def crs(self) -> CRS | None:
        """
        Return the coordinate reference system associated with the geometries.

        If multiple GeoDataFrames are present with differing CRS values, the CRS of the first
        GeoDataFrame that defines one is returned. If none of the stored geometries provide a
        CRS, the method returns None.

        Returns
        -------
        CRS or None
            The CRS of the first geometry with a defined coordinate reference system, or None
            if no CRS is available.
        """
        for geom in self.data.values():
            if geom.crs is not None:
                return geom.crs
        return None

    @property
    def _region_data(self) -> GeoDataFrame | None:
        """
        Return the region as the geometric union of all polygonal features stored in the data dictionary.

        Each GeoDataFrame in ``self.data`` is exploded to ensure that multi-part geometries are treated
        as individual components before computing the union. Only geometries with area contribute to the
        resulting region; non-area features such as points and lines are ignored.

        Returns
        -------
        geopandas.GeoDataFrame or None
            A GeoDataFrame containing the unified region geometry, or None if no polygonal features
            are available.
        """
        # Use the union of all geometries as region
        if len(self.data) == 0:
            return None

        # Flatten all geometries from all GeoDataFrames
        all_geoms = [
            geom
            for gdf in self.data.values()
            if not gdf.empty
            for geom in gdf.geometry.explode(index_parts=False)
            if geom is not None and not geom.is_empty and geom.geom_type == "Polygon"
        ]
        union_geom = gpd.GeoSeries(all_geoms, crs=self.crs).union_all()
        return gpd.GeoDataFrame(geometry=[union_geom], crs=self.crs)

    def set(self, geom: GeoDataFrame | GeoSeries, name: str):
        """Add data to the geom component.

        Arguments
        ---------
        geom: geopandas.GeoDataFrame or geopandas.GeoSeries
            New geometry data to add
        name: str
            Geometry name.
        """
        self._initialize()
        assert self._data is not None
        if name in self._data:
            logger.warning(f"Replacing geom: {name}")

        if isinstance(geom, GeoSeries):
            geom = cast(GeoDataFrame, geom.to_frame())

        # Verify if a geom is set to model crs and if not sets geom to model crs
        model_crs = self.model.crs or self.crs
        if model_crs and model_crs != geom.crs:
            geom.to_crs(model_crs.to_epsg(), inplace=True)
        self._data[name] = geom

    @hydromt_step
    def read(self, filename: str | None = None, **kwargs) -> None:
        r"""Read model geometries files at <root>/<filename>.

        Parameters
        ----------
        filename : str, optional
            Filename relative to model root. Should contain either a * wildcard character
            or a {name} placeholder to read multiple files into the data dictionary.
            All files matching the glob pattern defined by filename will be read.
            If a {name} placeholder is used, that name will be used as the key in the
            data dictionary.
            If no {name} placeholder is used, the filename without extension will be used
            as the key in the data dictionary.
            If None, the path that was provided at init will be used.
        **kwargs:
            Additional keyword arguments that are passed to the
            `geopandas.read_file` function.
        """
        self.root._assert_read_mode()
        self._initialize(skip_read=True)
        placeholder_filename = str(filename or self._filename).replace("*", "{name}")
        for path, name in _expand_wildcards_and_name_placeholder(
            placeholder_filename, self.root.path
        ).items():
            geom = cast(GeoDataFrame, gpd.read_file(path, **kwargs))
            logger.debug(f"Reading model file {name} at {path}.")
            self.set(geom=geom, name=name)

    @hydromt_step
    def write(
        self,
        filename: str | None = None,
        *,
        to_wgs84: bool = False,
        **kwargs,
    ) -> None:
        r"""Write model geometries to a vector file (by default GeoJSON) at <root>/<filename>.

        key-word arguments are passed to :py:meth:`geopandas.GeoDataFrame.to_file`

        Parameters
        ----------
        filename : str | None, optional
            filename relative to model root. should contain a {name} placeholder
            which will be used to determine the names/keys of the geometries.
            if None, the path that was provided at init will be used.
            Can be a relative path.
        to_wgs84: bool, optional
            If True, the geoms will be reprojected to WGS84(EPSG:4326) before they are written.
        **kwargs:
            Additional keyword arguments that are passed to the
            `geopandas.to_file` function.
        """
        self.root._assert_write_mode()

        if len(self.data) == 0:
            logger.info(
                f"{self.model.name}.{self.name_in_model}: No geoms data found, skip writing."
            )
            return
        placeholder_filename = str(filename or self._filename).replace("*", "{name}")
        for name, gdf in self.data.items():
            if len(gdf) == 0:
                logger.warning(
                    f"{self.model.name}.{self.name_in_model}: {name} is empty. Skipping..."
                )
                continue

            write_path = self.root.path / placeholder_filename.format(name=name)
            write_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(
                f"{self.model.name}.{self.name_in_model}: Writing geoms to {write_path}."
            )

            if to_wgs84 and (
                kwargs.get("driver") == "GeoJSON"
                or write_path.suffix.lower() == ".geojson"
            ):
                gdf.to_crs(epsg=4326, inplace=True)

            gdf.to_file(write_path, **kwargs)

    def test_equal(self, other: ModelComponent) -> tuple[bool, dict[str, str]]:
        """Test if two GeomsComponents are equal.

        Parameters
        ----------
        other: GeomsComponent
            The other GeomsComponent to compare with.

        Returns
        -------
        tuple[bool, dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_geoms = cast(GeomsComponent, other)
        for name, gdf in self.data.items():
            if name not in other_geoms.data:
                errors[name] = "Geom not found in other component."
            try:
                assert_geodataframe_equal(
                    gdf,
                    other_geoms.data[name],
                    check_like=True,
                    check_less_precise=True,
                )
            except AssertionError as e:
                errors[name] = str(e)

        return len(errors) == 0, errors



"""Implementations related to logging."""

import logging
import sys
from contextlib import contextmanager
from pathlib import Path

from hydromt import __version__

__all__ = [
    "initialize_logging",
    "set_log_level",
    "to_file",
]

_ROOT_LOGGER = logging.getLogger("hydromt")
_LOG_FORMAT = "%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s"
_DEFAULT_FORMATTER = logging.Formatter(_LOG_FORMAT)


def initialize_logging() -> None:
    """Initialize the hydromt root logger with a formatter, a log level and an optional console handler.

    This function should be called once, at the start of the program.
    For HydroMT, this is achieved by calling it in the `hydromt.__init__.py` file, so it is
    automatically called when importing the hydromt package.
    If the root logger has a handler before hydromt is imported, no new handler will be added.
    """
    logging.captureWarnings(True)
    _ROOT_LOGGER.setLevel(logging.INFO)
    if not _ROOT_LOGGER.hasHandlers():
        console = logging.StreamHandler(sys.stdout)
        console.setLevel(logging.INFO)
        console.setFormatter(_DEFAULT_FORMATTER)
        _ROOT_LOGGER.addHandler(console)


def set_log_level(log_level: int) -> None:
    """Set the log level of the hydromt root logger.

    This also affects all child loggers (e.g., ``hydromt.core``),
    unless they have their own log level explicitly set.

    Example
    -------
    .. code-block:: python
        from hydromt.log import set_log_level

        # Set log level to ERROR
        set_log_level(logging.ERROR)
    """
    _ROOT_LOGGER.setLevel(log_level)


def _add_filehandler(
    path: Path,
    *,
    log_level: int | None = None,
    formatter: logging.Formatter = _DEFAULT_FORMATTER,
    logger: logging.Logger = _ROOT_LOGGER,
) -> logging.FileHandler:
    """Add file handler to the hydromt root logger.

    Parameters
    ----------
    path : Path
        Path to the log file.
    log_level : int, optional
        Log level for the file handler, by default None (inherits from logger)
    formatter : logging.Formatter, optional
        Log formatter, by default _DEFAULT_FORMATTER
    logger : logging.Logger, optional
        Logger to add the file handler to, by default _ROOT_LOGGER

    Returns
    -------
    logging.FileHandler
        The created file handler.
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    filehandler = logging.FileHandler(path)
    filehandler.setFormatter(formatter)
    if log_level is not None:
        filehandler.setLevel(log_level)
    logger.addHandler(filehandler)
    if path.exists():
        logger.debug(f"Appending log messages to file {path}.")
    else:
        logger.debug(f"Writing log messages to new file {path}.")
    return filehandler


def log_version() -> None:
    _ROOT_LOGGER.info(f"HydroMT version: {__version__}")


@contextmanager
def to_file(
    path: Path,
    *,
    log_level: int | None = None,
    formatter: logging.Formatter = _DEFAULT_FORMATTER,
    logger: logging.Logger = _ROOT_LOGGER,
    append: bool = True,
):
    """Context manager that attaches a file handler for the duration of the block.

    Parameters
    ----------
    path : Path
        Path to the log file.
    log_level : int, optional
        Log level for the file handler, by default None (inherits from logger)
    formatter : logging.Formatter, optional
        Log formatter, by default _DEFAULT_FORMATTER
    logger : logging.Logger, optional
        Logger to add the file handler to, by default _ROOT_LOGGER
    append : bool, optional
        Whether to append to the log file if it exists, by default True.

    Example
    -------
    .. code-block:: python
        from hydromt import log
        from pathlib import Path
        log_file = Path.cwd() / "my_log.log"
        with log.to_file(log_file):
            # log messages in this block will be written to `log_file`
            logger = logging.getLogger("hydromt.my_module")
            logger.info("This is an info message.")
            logger.error("This is an error message.")

    """
    path.parent.mkdir(parents=True, exist_ok=True)
    if not append and path.exists():
        path.unlink()
    handler = _add_filehandler(
        path, log_level=log_level, formatter=formatter, logger=logger
    )
    try:
        yield
    finally:
        logger.removeHandler(handler)
        handler.close()



"""Implementation for the geodataset DataAdapter."""

import logging
from typing import Dict, List, Optional, Union, cast

import numpy as np
import pyproj
import xarray as xr

from hydromt._utils import (
    _has_no_data,
    _rename_vars,
    _set_metadata,
    _set_vector_nodata,
    _shift_dataset_time,
    _single_var_as_array,
)
from hydromt.data_catalog.adapters.adapter_utils import _slice_temporal_dimension
from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.gis.raster import GEO_MAP_COORD
from hydromt.typing import (
    Geom,
    Predicate,
    SourceMetadata,
    TimeRange,
    Variables,
)
from hydromt.typing.type_def import Number

logger = logging.getLogger(__name__)

__all__ = ["GeoDatasetAdapter"]


class GeoDatasetAdapter(DataAdapterBase):
    """DatasetAdapter for GeoDatasets."""

    def transform(
        self,
        ds: xr.Dataset,
        metadata: SourceMetadata,
        *,
        mask: Optional[Geom] = None,
        predicate: Predicate = "intersects",
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        single_var_as_array: bool = True,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[Union[xr.Dataset, xr.DataArray]]:
        """Return a clipped, sliced and harmonized RasterDataset.

        Parameters
        ----------
        ds : xr.Dataset
            input GeoDataset
        metadata : SourceMetadata
            source metadata
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        predicate : str, optional
            predicate to use for the mask filter, by default "intersects"
        variables : Optional[List[str]], optional
            variable filter, by default None
        time_range : Optional[TimeRange], optional
            filter start and end times, by default None
        single_var_as_array : bool, optional
            whether to return a xr.DataArray if only a single variable is present, by default True
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        Optional[Union[xr.Dataset, xr.DataArray]]
            The filtered and harmonized GeoDataset, or None if no data was available

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        ds = _rename_vars(ds, self.rename)
        ds = GeoDatasetAdapter._validate_spatial_coords(ds)
        ds = GeoDatasetAdapter._set_crs(ds, crs=metadata.crs)
        ds = _set_vector_nodata(ds, metadata)
        ds = _shift_dataset_time(dt=self.unit_add.get("time", 0), ds=ds)
        ds = GeoDatasetAdapter._apply_unit_conversion(
            ds,
            unit_mult=self.unit_mult,
            unit_add=self.unit_add,
        )
        ds = _set_metadata(ds, metadata=metadata)
        ds = GeoDatasetAdapter._slice_data(
            ds,
            variables=variables,
            mask=mask,
            predicate=predicate,
            time_range=time_range,
            handle_nodata=handle_nodata,
        )
        if ds is None:
            return None

        return _single_var_as_array(ds, single_var_as_array, variables)

    @staticmethod
    def _validate_spatial_coords(
        ds: Optional[xr.Dataset],
    ) -> Optional[xr.Dataset]:
        if GEO_MAP_COORD in ds.data_vars:
            ds = ds.set_coords(GEO_MAP_COORD)
        try:
            ds.vector.set_spatial_dims()
            idim = ds.vector.index_dim
            if idim not in ds:  # set coordinates for index dimension if missing
                ds[idim] = xr.IndexVariable(idim, np.arange(ds.dims[idim]))
            coords = [ds.vector.x_name, ds.vector.y_name, idim]
            coords = [item for item in coords if item is not None]
            ds = ds.set_coords(coords)
        except ValueError:
            raise ValueError("GeoDataset: No spatial geometry dimension found")
        return ds

    @staticmethod
    def _set_crs(
        ds: xr.Dataset,
        crs: Union[str, int, None] = None,
    ) -> Optional[xr.Dataset]:
        # set crs
        if ds.vector.crs is None and crs is not None:
            ds.vector.set_crs(crs)
        elif ds.vector.crs is None:
            raise ValueError("GeoDataset: CRS not defined in data catalog or data.")
        elif crs is not None and ds.vector.crs != pyproj.CRS.from_user_input(crs):
            logger.warning(
                "GeoDataset: CRS from data catalog does not match CRS of"
                " data. The original CRS will be used. Please check your data catalog."
            )
        return ds

    @staticmethod
    def _slice_data(
        ds: Union[xr.Dataset, xr.DataArray],
        variables: Optional[Variables] = None,
        mask: Optional[Geom] = None,
        predicate: Predicate = "intersects",
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[xr.Dataset]:
        """Filter the GeoDataset.

        Parameters
        ----------
        ds : Optional[Union[xr.Dataset, xr.DataArray]]
            input dataset
        variables : Optional[List[str]], optional
            variable filter, by default None
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        predicate : str, optional
            predicate to use for the mask filter, by default "intersects"
        time_range : Optional[TimeRange], optional
            filter start and end times, by default None

        Returns
        -------
        Optional[xr.Dataset]
            the filtered GeoDataSet

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE

        """
        if isinstance(ds, xr.DataArray):
            if ds.name is None:
                # dummy name, required to create dataset
                # renamed to variable in _single_var_as_array
                ds.name = "data"
            ds = ds.to_dataset()
        elif variables is not None:
            assert isinstance(ds, xr.Dataset)
            variables = cast(List, np.atleast_1d(variables).tolist())
            if len(variables) > 1 or len(ds.data_vars) > 1:
                mvars = [var not in ds.data_vars for var in variables]
                if any(mvars):
                    raise ValueError(f"GeoDataset: variables not found {mvars}")
                ds = ds[variables]

        if time_range is not None:
            ds = _slice_temporal_dimension(ds, time_range, handle_nodata=handle_nodata)
            if ds is None:
                return None
        if mask is not None:
            ds = GeoDatasetAdapter._slice_spatial_dimension(
                ds, mask=mask, predicate=predicate, handle_nodata=handle_nodata
            )
            if ds is None:
                return None

        return ds

    @staticmethod
    def _slice_spatial_dimension(
        ds: xr.Dataset,
        mask: Geom,
        predicate: Predicate,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[xr.Dataset]:
        bbox_str = ", ".join([f"{c:.3f}" for c in mask.total_bounds])
        epsg = mask.crs.to_epsg()
        logger.debug(f"Clip {predicate} [{bbox_str}] (EPSG:{epsg})")
        ds = ds.vector.clip_geom(mask, predicate=predicate)
        if _has_no_data(ds):
            exec_nodata_strat("No data left after spatial slicing.", handle_nodata)
            return None

        return ds

    @staticmethod
    def _apply_unit_conversion(
        ds: Optional[xr.Dataset],
        unit_mult: Dict[str, Number],
        unit_add: Dict[str, Number],
    ) -> Optional[xr.Dataset]:
        if ds is None:
            return None

        unit_names = list(unit_mult.keys()) + list(unit_add.keys())
        unit_names = [k for k in unit_names if k in ds.data_vars]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} variables.")
        for name in list(set(unit_names)):  # unique
            m = unit_mult.get(name, 1)
            a = unit_add.get(name, 0)
            da = ds[name]
            attrs = da.attrs.copy()
            nodata_isnan = da.vector.nodata is None or np.isnan(da.vector.nodata)
            # nodata value is explicitly set to NaN in case no nodata value is provided
            nodata = np.nan if nodata_isnan else da.vector.nodata
            data_bool = ~np.isnan(da) if nodata_isnan else da != nodata
            ds[name] = xr.where(data_bool, da * m + a, nodata)
            ds[name].attrs.update(attrs)  # set original attributes
        return ds



"""Utility functions for hydromt that have no other home."""

from hydromt._utils.caching import cache_vrt_tiles, copy_to_local
from hydromt._utils.dataset import (
    _rename_vars,
    _set_metadata,
    _shift_dataset_time,
    _single_var_as_array,
)
from hydromt._utils.deep_merge import _deep_merge
from hydromt._utils.dictionaries import _partition_dictionaries
from hydromt._utils.elevation import _elevation2rgba, _rgba2elevation
from hydromt._utils.format import CatalogDumper
from hydromt._utils.log import initialize_logging, to_file
from hydromt._utils.nodata import _has_no_data, _set_raster_nodata, _set_vector_nodata
from hydromt._utils.path import _make_config_paths_absolute, _make_config_paths_relative
from hydromt._utils.rgetattr import _rgetattr
from hydromt._utils.steps_validator import _validate_steps
from hydromt._utils.temp_env import temp_env
from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt._utils.uris import _is_valid_url, _strip_scheme

__all__ = [
    "CatalogDumper",
    "cache_vrt_tiles",
    "copy_to_local",
    "_deep_merge",
    "_rename_vars",
    "_set_metadata",
    "initialize_logging",
    "to_file",
    "_shift_dataset_time",
    "_single_var_as_array",
    "_partition_dictionaries",
    "_elevation2rgba",
    "_rgba2elevation",
    "_has_no_data",
    "_set_raster_nodata",
    "_set_vector_nodata",
    "_rgetattr",
    "_validate_steps",
    "_warn_on_unused_kwargs",
    "_is_valid_url",
    "_strip_scheme",
    "_make_config_paths_absolute",
    "_make_config_paths_relative",
    "temp_env",
]


class _classproperty(property):
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)



"""Drivers responsible for reading and writing data."""

from hydromt.data_catalog.drivers.base_driver import BaseDriver, DriverOptions
from hydromt.data_catalog.drivers.dataframe import DataFrameDriver, PandasDriver
from hydromt.data_catalog.drivers.dataset import DatasetDriver, DatasetXarrayDriver
from hydromt.data_catalog.drivers.geodataframe import (
    GeoDataFrameDriver,
    GeoDataFrameTableDriver,
    PyogrioDriver,
)
from hydromt.data_catalog.drivers.geodataframe.table_driver import (
    GeoDataFrameTableOptions,
)
from hydromt.data_catalog.drivers.geodataset import (
    GeoDatasetDriver,
    GeoDatasetVectorDriver,
    GeoDatasetXarrayDriver,
)
from hydromt.data_catalog.drivers.preprocessing import (
    harmonise_dims,
    remove_duplicates,
    round_latlon,
    to_datetimeindex,
)
from hydromt.data_catalog.drivers.raster.raster_dataset_driver import (
    RasterDatasetDriver,
)
from hydromt.data_catalog.drivers.raster.raster_xarray_driver import (
    RasterDatasetXarrayDriver,
)
from hydromt.data_catalog.drivers.raster.rasterio_driver import (
    RasterioDriver,
    RasterioOptions,
)
from hydromt.data_catalog.drivers.xarray_options import XarrayDriverOptions

__all__ = [
    "BaseDriver",
    "GeoDataFrameTableOptions",
    "DriverOptions",
    "DatasetDriver",
    "DatasetXarrayDriver",
    "DataFrameDriver",
    "GeoDataFrameDriver",
    "GeoDataFrameTableDriver",
    "GeoDatasetDriver",
    "GeoDatasetVectorDriver",
    "GeoDatasetXarrayDriver",
    "PandasDriver",
    "PyogrioDriver",
    "RasterDatasetDriver",
    "RasterDatasetXarrayDriver",
    "RasterioDriver",
    "RasterioOptions",
    "harmonise_dims",
    "remove_duplicates",
    "round_latlon",
    "to_datetimeindex",
    "XarrayDriverOptions",
]

# define hydromt driver entry points
# see also hydromt.driver group in pyproject.toml
__hydromt_eps__ = [
    "DatasetXarrayDriver",
    "GeoDataFrameTableDriver",
    "GeoDatasetVectorDriver",
    "GeoDatasetXarrayDriver",
    "PandasDriver",
    "PyogrioDriver",
    "RasterDatasetXarrayDriver",
    "RasterioDriver",
]



from functools import reduce

__all__ = ["_rgetattr"]


def _rgetattr(obj, attr, *args):
    """Recursive get attribute from object."""

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return reduce(_getattr, [obj] + attr.split("."))



"""Helper function to merge deeply nested dictionaries."""

from typing import Any, Dict


def _deep_merge(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
    """Merge deeply nested dictionaries.

    Given overlap in the leaves the right dictionary will take precedence.
    """
    merged = {}

    for k_left, v_left in left.items():
        merged[k_left] = v_left

    for k_right, v_right in right.items():
        if k_right in merged:
            v_left = merged[k_right]
            if isinstance(v_left, dict) and isinstance(v_right, dict):
                merged[k_right] = _deep_merge(v_left, v_right)
            else:
                merged[k_right] = v_right
        else:
            merged[k_right] = v_right

    return merged



"""Driver to read geodataframes using Pyogrio."""

import logging
from os.path import splitext
from pathlib import Path
from typing import Any, ClassVar

import geopandas as gpd
import pandas as pd
from pyogrio import read_dataframe, read_info, write_dataframe
from pyproj import CRS

from hydromt._utils.unused_kwargs import _warn_on_unused_kwargs
from hydromt.data_catalog.drivers.geodataframe.geodataframe_driver import (
    GeoDataFrameDriver,
)
from hydromt.error import NoDataStrategy, exec_nodata_strat
from hydromt.typing import Bbox, Geom, SourceMetadata

logger = logging.getLogger(__name__)


class PyogrioDriver(GeoDataFrameDriver):
    """
    Driver for GeoDataFrame using the pyogrio library: ``pyogrio``.

    Supports reading and writing files supported by the OGR library,
    including geopackage, shapefile, geojson and flatgeobuf.
    """

    name: ClassVar[str] = "pyogrio"
    supports_writing: ClassVar[bool] = True
    SUPPORTED_EXTENSIONS: ClassVar[set[str]] = {".gpkg", ".shp", ".geojson", ".fgb"}

    def read(
        self,
        uris: list[str],
        *,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        metadata: SourceMetadata | None = None,
        mask: Any = None,
        variables: str | list[str] | None = None,
    ) -> gpd.GeoDataFrame:
        """
        Read geospatial data using the pyogrio library into a GeoDataFrame.

        Supports formats such as GeoPackage, Shapefile, GeoJSON, and FlatGeobuf.
        Optionally applies spatial filtering through a bounding box derived
        from a provided mask.

        Parameters
        ----------
        uris : list[str]
            List of URIs to read data from. Only one file is supported per read operation.
        handle_nodata : NoDataStrategy, optional
            Strategy to handle missing or empty data. Default is NoDataStrategy.RAISE.
        metadata : SourceMetadata | None, optional
            Optional metadata object describing the dataset source (e.g. CRS).
        mask : Any, optional
            Optional geometry or GeoDataFrame used to spatially filter the data
            while reading.
        variables : str | list[str] | None, optional
            Optional list of columns to load from the dataset.

        Returns
        -------
        gpd.GeoDataFrame
            The loaded geospatial data.

        Raises
        ------
        ValueError
            If multiple URIs are provided.
        IOError
            If the source file contains no geometry column.

        Warning
        -------
        The `metadata` parameter is not used directly in this driver, but is included
        for consistency with the GeoDataFrameDriver interface.
        """
        _warn_on_unused_kwargs(self.__class__.__name__, {"metadata": metadata})

        if len(uris) > 1:
            raise ValueError(
                "DataFrame: Reading multiple files with the "
                f"{self.__class__.__name__} driver is not supported."
            )
        _uri = uris[0]
        if mask is not None:
            bbox = _bbox_from_file_and_mask(
                _uri, mask=mask, **self.options.get_kwargs()
            )
        else:
            bbox = None
        gdf: pd.DataFrame | gpd.GeoDataFrame = read_dataframe(
            _uri, bbox=bbox, columns=variables, **self.options.get_kwargs()
        )
        if not isinstance(gdf, gpd.GeoDataFrame):
            raise IOError(f"DataFrame from uri: '{_uri}' contains no geometry column.")

        if gdf.index.size == 0:
            exec_nodata_strat(
                f"No data from {self.name} driver for file uris: {', '.join(uris)}.",
                strategy=handle_nodata,
            )
            return None  # handle_nodata == ignore
        return gdf

    def write(
        self,
        path: Path | str,
        data: gpd.GeoDataFrame,
        *,
        write_kwargs: dict[str, Any] | None = None,
    ) -> Path:
        """
        Write a GeoDataFrame to disk using the pyogrio library.

        Supports writing to vector formats supported by the OGR library, including
        GeoPackage (`.gpkg`), Shapefile (`.shp`), GeoJSON (`.geojson`), and FlatGeobuf (`.fgb`).
        The file format is inferred from the file extension. If the extension is unsupported,
        it falls back to FlatGeobuf (`.fgb`).

        Parameters
        ----------
        path : Path | str
            Destination path or URI where the GeoDataFrame will be written.
            Supported extensions are `.gpkg`, `.shp`, `.geojson`, and `.fgb`.
        data : gpd.GeoDataFrame
            The GeoDataFrame to write.
        write_kwargs : dict[str, Any], optional
            Additional keyword arguments passed to `pyogrio.write_dataframe`. Default is None.

        Returns
        -------
        Path
            The path where the GeoDataFrame was written.

        Raises
        ------
        ValueError
            If the file extension cannot be determined or writing fails.
        """
        no_ext, ext = splitext(path)
        write_kwargs = write_kwargs or {}
        if ext not in self.SUPPORTED_EXTENSIONS:
            logger.warning(
                f"driver {self.name} has no support for extension {ext}"
                "switching to .fgb."
            )
            path = no_ext + ".fgb"

        write_dataframe(data, path, **write_kwargs)

        return Path(path)


def _bbox_from_file_and_mask(
    uri: str,
    mask: Geom,
    **kwargs,
) -> Bbox | None:
    """Create a bbox from the file metadata and mask given.

    Pyogrio's mask or bbox arguments require a mask or bbox in the same CRS as the data.
    This function takes the mask filter and crs of the input data
    and returns a bbox in the same crs as the data based on the input filters.

    Parameters
    ----------
    uri: str,
        URI of the data.
    mask: GeoDataFrame | GeoSeries | BaseGeometry
        mask to filter the data while reading.
    """
    source_crs = None
    if source_crs_str := read_info(uri, **kwargs).get("crs"):
        source_crs = CRS.from_user_input(source_crs_str)

    if not source_crs:
        logger.warning(
            f"Reading from uri: '{uri}' without CRS definition. Filtering with crs:"
            f" {mask.crs}, cannot compare crs."
        )
    elif mask.crs != source_crs:
        mask = mask.to_crs(source_crs)

    return tuple(mask.total_bounds)



"""Deferred file close handling to ensure data integrity during file operations."""

import logging
import shutil
from pathlib import Path

logger = logging.getLogger(__name__)

_MAX_CLOSE_ATTEMPTS = 2


class DeferredFileClose:
    """Class to handle deferred file closing operations."""

    def __init__(self, *, original_path: Path, temp_path: Path):
        self._original_path = original_path
        self._temp_path = temp_path
        self._close_attempts = 0

    def close(self) -> None:
        """Attempt to move the temporary file to the original path up to a maximum number of attempts."""
        while self._close_attempts < _MAX_CLOSE_ATTEMPTS:
            try:
                logger.debug(
                    f"Moving temporary file '{self._temp_path}' to destination '{self._original_path}'."
                )
                shutil.move(self._temp_path, self._original_path)
                return
            except PermissionError as e:
                self._close_attempts += 1
                logger.error(
                    f"Could not write to destination file {self._original_path} because the following error was raised: {e}"
                )
            except FileNotFoundError:
                logger.warning(
                    f"Could not find temporary file {self._temp_path}. It was likely already deleted by another component that updates the same dataset."
                )
                return

        # already tried to close this too many times
        logger.error(
            f"Max write attempts to file {self._original_path} exceeded. Skipping... "
            f"Instead, data was written to a temporary file: {self._temp_path}."
        )



"""Implementation for the RasterDatasetAdapter."""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, cast

import numpy as np
import pyproj
import xarray as xr
from pyproj import CRS

from hydromt._utils import (
    _has_no_data,
    _shift_dataset_time,
    _single_var_as_array,
)
from hydromt.data_catalog.adapters.adapter_utils import _slice_temporal_dimension
from hydromt.data_catalog.adapters.data_adapter_base import DataAdapterBase
from hydromt.error import NoDataException, NoDataStrategy, exec_nodata_strat
from hydromt.gis.raster import GEO_MAP_COORD
from hydromt.gis.raster_utils import _meridian_offset
from hydromt.typing import (
    Data,
    Geom,
    SourceMetadata,
    TimeRange,
    Variables,
)

logger = logging.getLogger(__name__)

__all__ = ["RasterDatasetAdapter"]


class RasterDatasetAdapter(DataAdapterBase):
    """Implementation for the RasterDatasetAdapter."""

    def transform(
        self,
        ds: xr.Dataset,
        metadata: SourceMetadata,
        *,
        mask: Optional[Geom] = None,
        align: Optional[bool] = None,
        variables: Optional[Variables] = None,
        time_range: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        single_var_as_array: bool = True,
        buffer: int = 0,
    ) -> xr.Dataset | xr.DataArray | None:
        """Filter and harmonize the input RasterDataset.

        Parameters
        ----------
        ds : xr.Dataset
            input RasterDataset
        metadata : SourceMetadata
            source metadata
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        variables : Optional[List[str]], optional
            variable filter, by default None
        time_range : Optional[TimeRange], optional
            filter start and end times, by default None
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE
        single_var_as_array : bool, optional
            whether to return a xr.DataArray if only a single variable is present, by default True
        buffer : int, optional
            Buffer around the `bbox` or `geom` area of interest expressed in resolution multiplicity,
            by default 0

        Returns
        -------
        Optional[xr.Dataset]
            The filtered and harmonized RasterDataset, or None if no data was available

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        # rename variables and parse data and attrs
        ds = self._rename_vars(ds)
        ds = self._validate_spatial_dims(ds)
        ds = self._set_crs(ds, metadata.crs)
        ds = self._set_nodata(ds, metadata)
        ds = _shift_dataset_time(dt=self.unit_add.get("time", 0), ds=ds)
        # slice data
        ds = RasterDatasetAdapter._slice_data(
            ds,
            variables,
            mask,
            align,
            time_range,
            buffer=buffer,
            handle_nodata=handle_nodata,
        )
        if ds is None:
            return None  # if handle_nodata ignore

        # uniformize data
        ds = self._apply_unit_conversions(ds)
        ds = self._set_metadata(ds, metadata)
        # return array if single var and single_var_as_array
        return _single_var_as_array(ds, single_var_as_array, variables)

    def _rename_vars(self, ds: Data) -> Data:
        rm = {k: v for k, v in self.rename.items() if k in ds}
        ds = ds.rename(rm)
        return ds

    def _validate_spatial_dims(self, ds: Data) -> Data:
        if GEO_MAP_COORD in ds.data_vars:
            ds = ds.set_coords(GEO_MAP_COORD)
        try:
            ds.raster.set_spatial_dims()
            # transpose dims to get y and x dim last
            x_dim = ds.raster.x_dim
            y_dim = ds.raster.y_dim
            ds = ds.transpose(..., y_dim, x_dim)
        except ValueError:
            raise ValueError(
                f"RasterDataset: No valid spatial coords found in data {self.path}"
            )
        return ds

    def _set_crs(self, ds: Data, crs: Optional[CRS]) -> Data:
        # set crs
        if ds.raster.crs is None and crs is not None:
            ds.raster.set_crs(crs)
        elif ds.raster.crs is None:
            raise ValueError("RasterDataset: CRS not defined in data catalog or data.")
        elif crs is not None and ds.raster.crs != pyproj.CRS.from_user_input(crs):
            logger.warning(
                "RasterDataset: CRS from data catalog does not match CRS "
                " of data. The original CRS will be used. Please check your catalog."
            )
        return ds

    @staticmethod
    def _slice_data(
        ds: Data,
        variables: Optional[List] = None,
        mask: Optional[Geom] = None,
        align: Optional[float] = None,
        time_range: Optional[TimeRange] = None,
        buffer: int = 0,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[xr.Dataset]:
        """Filter the RasterDataset.

        Parameters
        ----------
        ds : xr.Dataset
            input Dataset.
        variables : Optional[List[str]], optional
            variable filter, by default None
        mask : Optional[gpd.GeoDataFrame], optional
            mask to filter by geometry, by default None
        align : Optional[float], optional
            resolution to align the bounding box, by default None
        time_range : Optional[TimeRange], optional
            filter start and end times, by default None
        buffer : int, optional
            Buffer around the `bbox` or `geom` area of interest expressed in resolution multiplicity,
            by default 0
        handle_nodata : NoDataStrategy, optional
            how to handle no data being present in the result, by default NoDataStrategy.RAISE

        Returns
        -------
        Optional[Union[xr.Dataset, xr.DataArray]]
            The filtered and harmonized RasterDataset, or None if no data was available

        Raises
        ------
        ValueError
            if not all variables are found in the data
        NoDataException
            if no data in left after slicing and handle_nodata is NoDataStrategy.RAISE
        """
        if isinstance(ds, xr.DataArray):  # xr.DataArray has no variables
            if ds.name is None:
                # dummy name, required to create dataset
                # renamed to variable in _single_var_as_array
                ds.name = "data"
            ds = ds.to_dataset()
        elif variables is not None:  # xr.Dataset has variables
            variables = cast(List, np.atleast_1d(variables).tolist())
            if len(variables) > 1 or len(ds.data_vars) > 1:
                mvars = [var not in ds.data_vars for var in variables]
                if any(mvars):
                    raise NoDataException(f"RasterDataset: variables not found {mvars}")
                ds = ds[variables]

        if time_range is not None:
            ds = _slice_temporal_dimension(ds, time_range, handle_nodata=handle_nodata)
            if ds is None:
                return None

        if mask is not None:
            ds = RasterDatasetAdapter._slice_spatial_dimensions(
                ds, mask, buffer=buffer, align=align, handle_nodata=handle_nodata
            )
            if ds is None:
                return None

        return ds

    @staticmethod
    def _slice_spatial_dimensions(
        ds: Data,
        mask: Optional[Geom] = None,
        align: Optional[float] = None,
        buffer: int = 0,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ):
        # make sure bbox is in data crs
        bbox = None
        crs = ds.raster.crs
        epsg = crs.to_epsg()  # this could return None
        if mask is not None:
            bbox = mask.to_crs(crs).total_bounds
        # work with 4326 data that is defined at 0-360 degrees longtitude
        w, _, e, _ = ds.raster.bounds
        if epsg == 4326 and np.isclose(e - w, 360):  # allow for rounding errors
            ds = _meridian_offset(ds, bbox)

        # clip with bbox
        if bbox is not None:
            # check if bbox is fully covered
            bbox_str = ", ".join([f"{c:.3f}" for c in bbox])

            def _lt_or_close(a: float, b: float) -> bool:
                return np.isclose(a, b) or a < b

            w, s, e, n = ds.raster.bounds

            if not any(
                map(_lt_or_close, (w, s, bbox[2], bbox[3]), (bbox[0], bbox[1], e, n))
            ):
                logger.warning(
                    f"Dataset [{w}, {s}, {e}, {n}] does not fully cover bbox [{bbox_str}]"
                )

            logger.debug(f"Clip to [{bbox_str}] (epsg:{epsg}))")
            ds = ds.raster.clip_bbox(bbox, align=align, buffer=buffer)
            if _has_no_data(ds):
                exec_nodata_strat("No data left after bbox clipping.", handle_nodata)
                return None

        return ds

    def _apply_unit_conversions(self, ds: Data):
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in ds.data_vars]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} variables.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            da = ds[name]
            attrs = da.attrs.copy()
            nodata_isnan = da.raster.nodata is None or np.isnan(da.raster.nodata)
            # nodata value is explicitly set to NaN in case no nodata value is provided
            nodata = np.nan if nodata_isnan else da.raster.nodata
            data_bool = ~np.isnan(da) if nodata_isnan else da != nodata
            ds[name] = xr.where(data_bool, da * m + a, nodata)
            ds[name].attrs.update(attrs)  # set original attributes
            ds[name].raster.set_nodata(nodata)  # reset nodata in case of change

        return ds

    def _set_nodata(self, ds, metadata: SourceMetadata):
        """Parse and apply nodata values from the data catalog."""
        # set nodata value
        if nodata := metadata.nodata is not None:
            if not isinstance(nodata, dict):
                no_data_values: Dict[str, Any] = {
                    k: metadata.nodata for k in ds.data_vars.keys()
                }
            else:
                no_data_values: Dict[str, Any] = nodata
            for k in ds.data_vars:
                mv = no_data_values.get(k, None)
                if mv is not None and ds[k].raster.nodata is None:
                    ds[k].raster.set_nodata(mv)
        return ds

    def _set_metadata(self, ds, metadata: SourceMetadata):
        # unit attributes
        attrs = metadata.attrs
        for k in attrs:
            ds[k].attrs.update(attrs[k])
        # set meta data
        # exclude extent and attrs that are dict
        ds.attrs.update(
            metadata.model_dump(exclude=["attrs", "extent"], exclude_unset=True)
        )
        return ds



"""Model Region class."""

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, cast

import geopandas as gpd
from geopandas import GeoDataFrame
from pyproj import CRS

from hydromt.model.components.base import ModelComponent
from hydromt.writers import write_region

if TYPE_CHECKING:
    from hydromt.model import Model

logger = logging.getLogger(__name__)


class SpatialModelComponent(ModelComponent, ABC):
    """Base spatial model component for GIS components."""

    def __init__(
        self,
        model: "Model",
        *,
        region_component: Optional[str] = None,
        region_filename: str = "region.geojson",
    ) -> None:
        """
        Initialize a SpatialModelComponent.

        This component serves as a base class for components that are geospatial and
        require a region.

        To re-use in your won component, make sure you implement the `_region_data`
        property.

        Parameters
        ----------
        model: Model
            HydroMT model instance
        region_component: str, optional
            The name of the region component to use as reference for this component's
            region in case the region of this new component depends on the region of a
            different component. If None, the region will be set based on the
            `_region_data` property of the component itself.
        region_filename: str
            The path to use for writing the region data to a file.
            By default "region.geojson".
        """
        super().__init__(model)
        self._region_component = region_component
        self._region_filename = region_filename

    @property
    def bounds(self) -> Optional[Tuple[float, float, float, float]]:
        """Return the total bounds of the model region."""
        return self.region.total_bounds if self.region is not None else None

    @property
    def region(self) -> Optional[GeoDataFrame]:
        """Provide access to the underlying GeoDataFrame data of the model region."""
        region_from_reference = self._get_region_from_reference()
        return (
            region_from_reference
            if region_from_reference is not None
            else self._region_data
        )

    @property
    @abstractmethod
    def _region_data(self) -> Optional[GeoDataFrame]:
        """Implement this property in order to provide the region.

        This function will be called by the `region` property if no reference component is set.
        """
        raise NotImplementedError(
            "Property _region_data must be implemented in subclass."
        )

    @property
    def crs(self) -> Optional[CRS]:
        """Provide access to the CRS of the model region."""
        return self.region.crs if self.region is not None else None

    def write_region(
        self,
        filename: Optional[str] = None,
        *,
        to_wgs84: bool = False,
        to_file_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Write the model region to file.

        The region is an auxiliary file that is often not required by the model,
        but can be useful for getting data from the data catalog.
        Plugin implementors may choose to write this file on write for a specific component.

        Parameters
        ----------
        filename : str, optional
            The filename to write the region to. If None, the filename provided at initialization is used.
        to_wgs84 : bool
            If True, the region is reprojected to WGS84 before writing.
        to_file_kwargs: dict, optional
            Additional keyword arguments passed to the `geopandas.GeoDataFrame.to_file` function.
        """
        self.root._assert_write_mode()
        if self._region_component is not None:
            logger.info(
                "Region is a reference to another component. Skipping writing..."
            )
            return

        if self.region is None:
            logger.warning(
                f"{self.model.name}.{self.name_in_model}: No region data available to write."
            )
            return

        filename = filename or self._region_filename
        full_path = self.root.path / filename
        full_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(
            f"{self.model.name}.{self.name_in_model}: Writing region to {full_path}."
        )

        write_region(
            self.region,
            file_path=full_path,
            to_wgs84=to_wgs84,
            to_file_kwargs=to_file_kwargs,
        )

    def test_equal(self, other: ModelComponent) -> Tuple[bool, Dict[str, str]]:
        """Test if two components are equal.

        Parameters
        ----------
        other : ModelComponent
            The component to compare against.

        Returns
        -------
        tuple[bool, Dict[str, str]]
            True if the components are equal, and a dict with the associated errors per property checked.
        """
        eq, errors = super().test_equal(other)
        if not eq:
            return eq, errors
        other_region = cast(SpatialModelComponent, other)

        try:
            # empty components are still equal
            if self.region is not None or other.region is not None:
                gpd.testing.assert_geodataframe_equal(
                    self.region.copy(),
                    other_region.region.copy(),
                    check_like=True,
                    check_less_precise=True,
                )
        except AssertionError as e:
            errors["data"] = str(e)

        return len(errors) == 0, errors

    def _get_region_from_reference(self) -> Optional[gpd.GeoDataFrame]:
        if self._region_component is not None:
            region_component = cast(
                SpatialModelComponent, self.model.get_component(self._region_component)
            )
            if region_component is None:
                raise ValueError(
                    f"Unable to find the referenced region component: '{self._region_component}'"
                )
            if region_component.region is None:
                raise ValueError(
                    f"Unable to get region from the referenced region component: '{self._region_component}'"
                )
            return region_component.region
        return None



"""Functions for extreme value analysis."""

import math as math
from typing import Optional

import dask
import numpy as np
import xarray as xr
from numba import njit
from numpy.typing import NDArray
from scipy import stats

__all__ = [
    "eva",
    "get_peaks",
    "get_return_value",
    "fit_extremes",
]

_RPS: NDArray[np.int_] = np.array([2, 5, 10, 25, 50, 100, 250, 500])  # years
_DISTS = {
    "POT": ["exp", "gpd"],
    "BM": ["gumb", "gev"],
}

_DEFAULT_PERIOD_STR = "365.25D"

## high level methods


def eva(
    da: xr.DataArray,
    ev_type: str = "BM",
    min_dist: int = 0,
    qthresh: float = 0.9,
    period: str = _DEFAULT_PERIOD_STR,
    min_sample_size: int = 0,
    distribution: Optional[str] = None,
    rps: NDArray[np.int_] = _RPS,
    criterium: str = "AIC",
) -> xr.Dataset:
    """Return Extreme Value Analysis.

    Extreme value analysis based on block maxima (BM) or Peaks Over Threshold (POT).
    The method selects the peaks, fits a distribution and calculates return values for
    provided return periods.

    Parameters
    ----------
    da : xr.DataArray
        Timeseries data, must have a regular spaced 'time' dimension.
    ev_type: {"POT", "BM"}
        Peaks over threshold (POT) or block maxima (BM) peaks, by default "BM"
    period : str, optional
        Period string, by default "365.25D". See pandas.Timedelta for options.
    min_dist : int, optional
        Minimum distance between peaks measured in time steps, by default 0
    qthresh : float, optional
        Quantile threshold used with peaks over threshold method, by default 0.9
    min_sample_size : int, optional
        Minimum number of finite values in a valid block, by default 0. Peaks of
        invalid blocks are set to NaN.
    distribution : str, optional
        Short name of distribution. If None (default) the optimal block maxima
        distribution ("gumb" or "gev" for BM and "exp" or "gpd" for POT) is selected
        based on `criterium`.
    rps : np.ndarray, optional
        Array of return periods, by default [2, 5, 10, 25, 50, 100, 250, 500]
    criterium: {'AIC', 'AICc', 'BIC'}
        Selection criterium, by default "AIC"

    Returns
    -------
    xr.Dataset
        Dataset with peaks timeseries, distribution name and parameters
        and return values.
    """
    da_peaks = get_peaks(
        da,
        ev_type=ev_type,
        min_dist=min_dist,
        qthresh=qthresh,
        period=period,
        min_sample_size=min_sample_size,
    )
    # fit distribution using lmom
    da_params = fit_extremes(
        da_peaks, ev_type=ev_type, distribution=distribution, criterium=criterium
    )
    # get return values
    da_rps = get_return_value(da_params, rps=rps)
    # combine data
    return xr.merge([da_peaks, da_params, da_rps])


def get_peaks(
    da: xr.DataArray,
    ev_type: str = "BM",
    min_dist: int = 0,
    qthresh: float = 0.9,
    period: str = "year",
    min_sample_size: int = 0,
    time_dim: str = "time",
) -> xr.DataArray:
    """Return peaks from time series.

    Return the timeseries with all but the peak values set to NaN.

    For block maxima (BM) peaks, peaks are determined by finding the maximum within
    each period and then ensuring a minimum distance (min_dist) and minimum sample size
    (min_sample_size) per period.

    For Peaks Over Threshold (POT), peaks are determined solely based on the minimum
    distance between peaks.

    The average interarrival time (extreme_rates) is calculated by dividing the number
    of peaks by the total duration of the timeseries and converting it to a yearly
    rate.

    Parameters
    ----------
    da : xr.DataArray
        Timeseries data, must have a regular spaced 'time' dimension.
    ev_type : {"POT", "BM"}
        Peaks over threshold (POT) or block maxima (BM) peaks, by default "BM"
    min_dist : int, optional
        Minimum distance between peaks measured in time steps, by default 0
    qthresh : float, optional
        Quantile threshold used with peaks over threshold method, by default 0.9
    period : {'year', 'month', 'quarter', pandas.Timedelta}, optional
        Period string, by default "year".
    min_sample_size : int, optional
        Minimum number of finite values in a valid block, by default 0. Peaks of
        invalid bins are set to NaN.

    Returns
    -------
    xr.DataArray
        Timeseries data with only peak values, all other values are set to NaN.
        Average interarrival time calculated based on the average number of peaks
        per year and stored as "extreme_rates"
    """
    if not (0 < qthresh < 1.0):
        raise ValueError("Quantile 'qthresh' should be between (0,1)")
    if time_dim not in da.dims:
        raise ValueError(f"Input array should have a '{time_dim}' dimension")
    if ev_type.upper() not in _DISTS.keys():
        raise ValueError(
            f"Unknown ev_type {ev_type.upper()}, select from {_DISTS.keys()}."
        )
    bins = None
    nyears = (da[time_dim][-1] - da[time_dim][0]).dt.days / 365.2425
    if period in ["year", "quarter", "month"] and ev_type.upper() == "BM":
        bins = getattr(da[time_dim].dt, period).values
    elif ev_type.upper() == "BM":
        tstart = da[time_dim].resample(time=period, label="left").first()
        bins = tstart.reindex_like(da, method="ffill").values.astype(float)
    else:
        min_sample_size = 0  # min_sample_size not used for POT

    def func(x):
        return local_max_1d(
            x, min_dist=min_dist, bins=bins, min_sample_size=min_sample_size
        )

    duck = dask.array if isinstance(da.data, dask.array.Array) else np
    lmax = duck.apply_along_axis(func, da.get_axis_num(time_dim), da.data)
    # apply POT threshold
    peaks = da.where(lmax)
    if ev_type.upper() == "POT":
        peaks = da.where(peaks > da.quantile(qthresh, dim=time_dim))
    # get extreme rate per year
    da_rate = np.isfinite(peaks).sum(time_dim) / nyears
    peaks = peaks.assign_coords({"extremes_rate": da_rate})
    peaks.name = "peaks"
    return peaks


def get_return_value(
    da_params: xr.DataArray,
    rps: NDArray[np.int_] = _RPS,
    extremes_rate: float = 1.0,
) -> xr.DataArray:
    """Return return value based on EVA.

    Return return values based on a fitted extreme value distribution using the
    :py:meth:`fit_extremes` method based on the scipy inverse survival function (isf).

    Parameters
    ----------
    da_params : xr.DataArray
        Short name and parameters of extreme value distribution,
        see also :py:meth:`fit_extremes`.
    rps : np.ndarray, optional
        Array of return periods in years, by default
        [1.5, 2, 5, 10, 20, 50, 100, 200, 500]
    extremes_rate : float, optional
        Average number of peaks per period, by default 1.0
        Only used if extremes_rate is not provided as coordinate in da_params.

    Returns
    -------
    xr.DataArray
        Return values
    """

    def _return_values_1d(p, r, d, rps=rps):
        if np.isnan(p).all():
            return np.full(rps.size, np.nan)
        if d == "gumb" and len(p) == 3:
            p = p[1:]
        return _get_return_values(p, d, rps=rps, extremes_rate=r)

    if isinstance(rps, list):
        rps = np.asarray(rps)
    elif not isinstance(rps, np.ndarray):
        raise ValueError("rps should be a list or numpy array")
    if "dparams" not in da_params.dims:
        raise ValueError("da_params should have a 'dparams' dimension")
    if "distribution" not in da_params.coords:
        raise ValueError("da_params should have a 'distribution' coordinate")
    distributions = da_params["distribution"].load()
    if "extremes_rate" in da_params.coords:
        extremes_rate = da_params["extremes_rate"].load()
    elif isinstance(extremes_rate, (int, float)):
        extremes_rate = np.full(distributions.shape, extremes_rate)
    elif extremes_rate.shape != distributions.shape:
        raise ValueError(
            "extremes_rate should be a scalar or have the same shape as distribution"
        )

    if da_params.ndim == 1:  # fix case of single dim
        da_params = da_params.expand_dims("index")
    da_rvs = xr.apply_ufunc(
        _return_values_1d,
        da_params,
        extremes_rate,
        distributions,
        input_core_dims=(["dparams"], [], []),
        output_core_dims=[["rps"]],
        vectorize=True,
        dask="parallelized",
        dask_gufunc_kwargs=dict(output_sizes={"rps": rps.size}),
        output_dtypes=[float],
    )
    da_rvs["rps"] = xr.IndexVariable("rps", rps)
    da_rvs.name = "return_values"
    return da_rvs.squeeze()


def fit_extremes(
    da_peaks: xr.DataArray,
    distribution: Optional[str] = None,
    ev_type: str = "BM",
    criterium: str = "AIC",
    time_dim: str = "time",
) -> xr.DataArray:
    """Return distribution fit from extremes.

    Return the fitted parameters of the extreme value `distribution` based on
    the lmoments method. If no distribution name is provided `distribution=None`,
    the optimal distribution is selected based on `criterium` from a list
    of distributions associated with `ev_type`.

    Block maximum distributions: gumbel ("gumb") and general extreme value ("gev").
    Peak over threshold distributions: exponential ("exp") and general pareto
    distribution ("gdp").

    Parameters
    ----------
    da_peaks : xr.DataArray
        Timeseries data with only peak values, any other values are set to NaN.
        The DataArray should contain as coordinate `extreme_rates` indicating
        the yearly rate of extreme events. If not provided, this value is set to 1.0
    distribution: {'gev', 'gpd', 'gumb', 'exp'}, optional
        Short distribution name. If None (default) the optimal distribution
        is calculated based on `criterium`
    ev_type : {"POT", "BM"}
        Peaks over threshold (POT) or block maxima (BM) peaks, by default "BM"
    criterium: {'AIC', 'AICc', 'BIC'}
        Selection criterium, by default "AIC"

    Returns
    -------
    xr.DataArray
        Parameters and short name of optimal extreme value distribution.
    """
    distributions = _DISTS.get(ev_type.upper(), None)
    if distribution is not None:
        if isinstance(distribution, str):
            distributions = [distribution]
        elif not isinstance(distribution, list):
            raise ValueError(
                f"distribution should be a string or list, got {type(distribution)}"
            )
    elif ev_type.upper() not in _DISTS:
        raise ValueError(
            f"Unknown ev_type {ev_type.upper()}, select from {_DISTS.keys()}."
        )

    def _fitopt_1d(x, distributions=distributions, criterium=criterium):
        if np.isnan(x).all():
            return np.array([np.nan, np.nan, np.nan, -1])
        params, d = lmoment_fitopt(x, distributions=distributions, criterium=criterium)
        if len(params) == 2:
            params = np.concatenate([[0], params])
        # trick to include distribution name; -1 if too little data to fit -> NA
        idist = distributions.index(d) if d in distributions else -1
        return np.concatenate([params, [idist]])

    if da_peaks.ndim == 1:  # fix case of single dim
        da_peaks = da_peaks.expand_dims("index")
    da_params = xr.apply_ufunc(
        _fitopt_1d,
        da_peaks,
        input_core_dims=[[time_dim]],
        output_core_dims=[["dparams"]],
        dask_gufunc_kwargs=dict(output_sizes={"dparams": 4}),
        vectorize=True,
        dask="parallelized",
        output_dtypes=[float],
    )
    # split output
    idist = da_params.isel(dparams=-1).values.astype(int)
    distributions = np.atleast_1d(np.array(distributions + ["NA"])[idist])
    da_params = da_params.isel(dparams=slice(0, -1))
    da_params.name = "parameters"
    # add coordinates
    dist_dims = list([d for d in da_params.dims if d != "dparams"])
    coords = dict(
        dparams=xr.IndexVariable("dparams", ["shape", "loc", "scale"]),
        distribution=xr.DataArray(dims=dist_dims, data=distributions),
    )
    # forward extremes_rate if provided
    if "extremes_rate" in da_peaks.coords:
        coords["extremes_rate"] = da_peaks["extremes_rate"]
    da_params = da_params.assign_coords(coords)
    return da_params.squeeze()


## PEAKS
@njit
def local_max_1d(
    arr: NDArray[np.number],
    bins: Optional[NDArray[np.int_]] = None,
    min_dist: int = 0,
    min_sample_size: int = 0,
) -> NDArray[np.bool_]:
    """Return boolean index of local maxima in `arr` which are `min_dist` apart.

    Parameters
    ----------
    arr : np.ndarray
        1D time series
    bins : np.ndarray, optional
        1D array of with uniquely numbered bins (blocks), by default None.
        If provided only the largest peak per block is flagged.
    min_dist : int, optional
        Minimum distance between peaks, by default 0
    min_sample_size : int, optional
        minimum number of samples per block, by default 0

    Returns
    -------
    np.ndarray
        boolean index of local maxima
    """
    a0 = arr[0]
    amax = -np.inf  # peak value
    imax = -min_dist  # peak index
    bsize = 0
    min_sample_size = 0 if bins is None else min_sample_size
    up = False  # sign of difference between subsequent values
    out = np.array([bool(0) for _ in range(arr.size)], np.bool_)
    for i in range(arr.size):
        a1 = arr[i]
        if not np.isfinite(a1):
            a0 = a1
            continue
        dd = i - 1 - imax  # distance to previous peak
        if (imax > 0) and (
            (bins is None and dd == (min_dist + 1))
            or (bins is not None and bins[i - 1] != bins[imax] and dd > min_dist)
        ):
            if bsize >= min_sample_size:
                out[imax] = True
            amax = -np.inf
            bsize = 0
        if up and a1 < a0 and a0 > amax:  # peak
            imax = i - 1
            amax = a0
        if a1 < a0:
            up = False
        elif a1 > a0:
            up = True
        bsize += 1
        a0 = a1
    if imax > 0 and bsize >= min_sample_size:
        out[imax] = True
    return out


## LINK TO SCIPY.STATS


def get_dist(distribution: str) -> stats.rv_continuous:
    """Return scipy.stats distribution."""
    _DISTS = {
        "gev": "genextreme",
        "gpd": "genpareto",
        "gumb": "gumbel_r",
        "exp": "genpareto",
    }
    _scipy_dist_name = _DISTS.get(distribution, distribution)
    dist = getattr(stats, _scipy_dist_name, None)
    if dist is None:
        raise ValueError(f'Distribution "{_scipy_dist_name}" not found in scipy.stats.')
    return dist


def get_frozen_dist(params, distribution: str):
    """Return frozen distribution.

    Returns scipy.stats frozen distribution, i.e.: with set parameters.
    """
    return get_dist(distribution)(*params[:-2], loc=params[-2], scale=params[-1])


## STATS


def _aic(x, params, distribution: str):
    """Return Akaike Information Criterion for a frozen distribution."""
    k = len(params)
    nll = get_frozen_dist(params, distribution).logpdf(x).sum()
    aic = 2 * k - 2 * nll
    return aic


def _aicc(x, params, distribution: str):
    """Return AICC.

    Return Akaike Information Criterion with correction for small sample size
    for a frozen distribution.
    """
    k = len(params)
    aic = _aic(x, params, distribution)
    aicc = aic + ((2 * k) ** 2 + 2 * k) / (len(x) - k - 1)
    return aicc


def _bic(x, params, distribution: str):
    """Return Bayesian Information Criterion for a frozen distribution."""
    k = len(params)
    nll = get_frozen_dist(params, distribution).logpdf(x).sum()
    bic = k * np.log(len(x)) - 2 * nll
    return bic


## TRANSFORMATIONS


def _get_return_values(
    params, distribution: str, rps: NDArray[np.int_] = _RPS, extremes_rate=1.0
):
    q = 1 / rps / extremes_rate
    return get_frozen_dist(params, distribution).isf(q)


def _get_return_periods(x, a=0.0, extremes_rate=1.0):
    assert np.all(np.isfinite(x))
    b = 1.0 - 2.0 * a
    ranks = (len(x) + 1) - stats.rankdata(x, method="average")
    freq = ((ranks - a) / (len(x) + b)) * extremes_rate
    rps = 1 / freq
    return rps


## CONFIDENCE INTERVALS


def lmoment_ci(x, distribution, nsample=1000, alpha=0.9, rps=_RPS, extremes_rate=1.0):
    q = 1 / rps / extremes_rate
    dist = get_dist(distribution)

    def func(x, distribution=distribution, q=q):
        p = lmoment_fit(x, distribution)
        return dist.isf(q, *p[:-2], loc=p[-2], scale=p[-1])

    x_sample = np.random.choice(x, size=[nsample, x.size], replace=True)
    xrv = np.apply_along_axis(func, 1, x_sample)

    percentiles = np.array([(1 - alpha) / 2, 1 - (1 - alpha) / 2]) * 100
    return np.percentile(xrv, percentiles, axis=0)


## PLOTS


def plot_return_values(
    x: xr.DataArray,
    params: xr.DataArray,
    distribution: str,
    x_scale: Optional[str] = "gumbel",
    ax=None,
    color: Optional[str] = "k",
    a: Optional[float] = 0,
    alpha: Optional[float] = 0.9,
    nsample: int = 1000,
    rps: NDArray[np.int_] = _RPS,
    extremes_rate: float = 1.0,
):
    """Return figure of EVA fit and empirical data.

    Parameters
    ----------
    x : xr.DataArray
        Timeseries data with only peak values, all other values are set to NaN
    params : xr.DataArray
        Short name and parameters of extreme value distribution,
        see also :py:meth:`fit_extremes`.
    distribution : str
        Name of distribution used for fit. Can be set to `empirical` if no fit is
        performed
    x_scale : str, optional
        transformation of the x-axis for the figure
    ax :

    color : str, optional
        Color of the line. By default colord is black 'k'.
    a : float, optional
        Float for the Gringorten position. By default a = 0
    alpha : float, optional
        alpha value for the confidence interval of the EV fit. By default, alpha = 0.9
    nsample : int
        number of samples used to calculate the confidence interval. By default,
        nsample = 1000
    rps : np.ndarray, optional
        Specific return periods to highlight as points in the plot. By default
        [2, 5, 10, 25, 50, 100, 250, 500] years
    extremes_rate : float.
        Average interarrival time calculated based on the average number of peaks
        per year. For annual maxima, extremes_rate = 1

    Returns
    -------
    ax.figure
        Figure of EVA fit and empirical data
    """
    import matplotlib.pyplot as plt

    rvs_obs = np.sort(x[np.isfinite(x)])
    if distribution != "empirical":
        params = params[-2:] if distribution == "gumb" else params
        rvs_sim = _get_return_values(
            params, distribution, rps=rps, extremes_rate=extremes_rate
        )
    else:
        rvs_sim = rvs_obs
    rps_obs = _get_return_periods(rvs_obs, a=a, extremes_rate=extremes_rate)

    if x_scale == "gumbel":
        xsim = -np.log(-np.log(1.0 - 1.0 / rps))
        xobs = -np.log(-np.log(1.0 - 1.0 / rps_obs))
    else:
        xsim = rps
        xobs = rps_obs

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(10, 6))
    ax.plot(xobs, rvs_obs, color=color, marker="o", label="plot position", lw=0)

    if distribution != "empirical":
        ax.plot(
            xsim, rvs_sim, color=color, ls="--", label=f"{distribution.upper()} fit"
        )

    if alpha is not None and nsample > 0 and distribution != "empirical":
        urvs = lmoment_ci(
            x,
            distribution,
            nsample=nsample,
            alpha=alpha,
            rps=rps,
            extremes_rate=extremes_rate,
        )
        ax.plot(
            xsim,
            urvs[0, :],
            color=color,
            ls=":",
            label=f"conf. interval (alpha = {alpha:.2f})",
        )
        ax.plot(
            xsim,
            urvs[1, :],
            color=color,
            ls=":",
        )

    ax.legend(loc="upper left")

    ymin = 0
    ymax = np.nanmax([np.nanmax(rvs_obs), np.nanmax(rvs_sim), 0])
    ymax = ymax * 1.1
    ax.set_ylim(ymin, ymax)

    ax.set_ylabel("Return value")
    if x_scale == "gumbel":
        ax.set_xticks(xsim)
        ax.set_xticklabels(rps)
    ax.set_xlabel("Return period")
    ax.grid()

    return ax


## LMOMENTS FITTING
# credits Ferdinand Diermanse


def lmoment_fitopt(x, distributions=None, criterium="AIC"):
    """Return parameters based on lmoments fit.

    Lmomentfit routine derive parameters of a distribution function
    based on lmoments. The distribution selection is based on the
    AIC criterium.

    Based on the theory of Hosking and Wallis (1997) Appendix A.

    Parameters
    ----------
    X: 1D array of float
        data series
    distributions: iterable of {'gev', 'gpd', 'gumb', 'exp'}
        iterable of distribution names
    criterium: {'AIC', 'AICc', 'BIC'}
        distrition selection criterium

    Returns
    -------
    params: ndarray of float
        array of distribution parameters
    distribution: str
        selected distribution
    """
    fgof = {"AIC": _aic, "AICC": _aicc, "BIC": _bic}.get(criterium.upper())
    # make sure the timeseries does not contain NaNs
    x = x[~np.isnan(x)]

    # derive first four L-moments from data
    lmom = get_lmom(x, 4)
    # derive parameters of distribution function
    params = {}
    gof_values = []
    for distribution in distributions:
        params[distribution] = _lmomentfit(lmom, distribution)
        gof_values.append(fgof(x, params[distribution], distribution))
    distribution = distributions[np.argmin(gof_values)]

    return params[distribution], distribution


def lmoment_fit(x, distribution):
    """Return parameters based on lmoments.

    Lmomentfit routine derive parameters of a distribution function
    based on lmoments.

    Based on the theory of Hosking and Wallis (1997) Appendix A.

    Parameters
    ----------
    X: 1D array of float
        data series
    distribution: {'gev', 'gpd', 'gumb', 'exp'}
        Short name of distribution function to be fitted.

    Returns
    -------
    params: ndarray of float
        array of distribution parameters
    lambda: 1D array of float
        vector of (nmom) L-moments
    """
    # make sure the timesiries does not contain NaNs
    x = x[~np.isnan(x)]

    # derive first four L-moments from data
    lmom = get_lmom(x, 4)

    # derive parameters of distribution function
    params = _lmomentfit(lmom, distribution)

    return params


def _lmomentfit(lmom, distribution):
    """Return parameters based on lmoments.

    Lmomentfit routine to derive parameters of a distribution function
    based on given lmoments.

    Based on the theory of Hosking and Wallis (1997) Appendix A.

    Parameters
    ----------
    lmom: 1D array of float
        l-moments, derived from data
    distribution: {'gev', 'gpd', 'gumb', 'exp'}
        Short name of distribution function to be fitted.

    Returns
    -------
    params: ndarray of float
        array of distribution parameters
    """
    # l-moment ratios from l-moments
    tau3 = lmom[2] / lmom[1]  # tau3 is in L-SK

    # derive parameters for selected distribution
    if distribution in ["gev", "genextreme"]:
        c1 = 2.0 / (3.0 + tau3) - np.log(2.0) / np.log(3.0)
        k1 = 7.859 * c1 + 2.9554 * (c1**2.0)
        s1 = (lmom[1] * k1) / ((1.0 - 2.0 ** (-k1)) * math.gamma(1.0 + k1))
        m1 = lmom[0] - (s1 / k1) * (1.0 - math.gamma(1.0 + k1))
        params = (k1, m1, s1)
    elif distribution in ["gumb", "gumbel_r"]:
        s1 = lmom[1] / np.log(2.0)
        m1 = lmom[0] - 0.5772 * s1
        params = (m1, s1)
    elif distribution in ["gpd", "genpareto"]:
        k1 = (1 - 3 * tau3) / (1 + tau3)
        s1 = (1 + k1) * (2 + k1) * lmom[1]
        m1 = lmom[0] - (2 + k1) * lmom[1]
        params = (-k1, m1, s1)
    elif distribution in ["exp", "genexpon"]:
        k1 = 1e-8
        s1 = (1 + k1) * (2 + k1) * lmom[1]
        m1 = lmom[0] - (2 + k1) * lmom[1]
        params = (0.0, m1, s1)
    else:
        raise ValueError("Unknown distribution")

    return params


def legendre_shift_poly(n):
    """Shifted Legendre polynomial.

    Based on recurrence relation
        (n + 1)Pn+1 (x) - (1 + 2 n)(2 x - 1)Pn (x) + n Pn-1 (x) = 0

    Given nonnegative integer n, compute the shifted Legendre polynomial P_n.
    Return the result as a vector whose mth element is the coefficient of x^(n+1-m).
    polyval(legendre_shift_poly(n),x) evaluates P_n(x).
    """
    if n == 0:
        pk = 1
    elif n == 1:
        pk = [2, -1]
    else:
        pkm2 = np.zeros(n + 1)
        pkm2[-1] = 1
        pkm1 = np.zeros(n + 1)
        pkm1[-1] = -1
        pkm1[-2] = 2

        for k in range(2, n + 1):
            pk = np.zeros(n + 1)

            for e in range(n - k + 1, n + 1):
                pk[e - 1] = (
                    (4 * k - 2) * pkm1[e]
                    + (1 - 2 * k) * pkm1[e - 1]
                    + (1 - k) * pkm2[e - 1]
                )

            pk[-1] = (1 - 2 * k) * pkm1[-1] + (1 - k) * pkm2[-1]
            pk = pk / k

            if k < n:
                pkm2 = pkm1
                pkm1 = pk

    return pk


def get_lmom(x, nmom=4):
    """Compute L-moments for a data series.

    Based on calculation of probability weighted moments and the coefficient
    of the shifted Legendre polynomial.

    lmom by Kobus N. Bekker, 14-09-2004

    Parameters
    ----------
    x: 1D array of float
        data series
    nmom: int
        number of L-Moments to be computed, by default 4.

    Returns
    -------
    lmom: 1D array of float
        vector of (nmom) L-moments
    """
    n = len(x)
    xs = np.sort(x, axis=0)
    bb = np.zeros(nmom - 1)
    ll = np.zeros(nmom - 1)
    b0 = xs.mean(axis=0)

    for r in range(1, nmom):
        num1 = np.kron(np.ones((r, 1)), np.arange(r + 1, n + 1))
        num2 = np.kron(np.ones((n - r, 1)), np.arange(1, r + 1)).T
        num = np.prod(num1 - num2, axis=0)

        den = np.prod(np.kron(np.ones((1, r)), n) - np.arange(1, r + 1))
        bb[r - 1] = (((num / den) * xs[r:n]).sum()) / n

    B = np.concatenate([np.array([b0]), bb.T])[::-1]

    for i in range(1, nmom):
        spc = np.zeros(len(B) - (i + 1))
        coeff = np.concatenate([spc, legendre_shift_poly(i)])
        ll[i - 1] = np.sum(coeff * B)

    lmom = np.concatenate([np.array([b0]), ll.T])

    return lmom



"""Implementations of the core ModelComponents."""

from hydromt.model.components.base import ModelComponent
from hydromt.model.components.config import ConfigComponent
from hydromt.model.components.datasets import DatasetsComponent
from hydromt.model.components.geoms import GeomsComponent
from hydromt.model.components.grid import GridComponent
from hydromt.model.components.mesh import MeshComponent
from hydromt.model.components.spatial import SpatialModelComponent
from hydromt.model.components.spatialdatasets import SpatialDatasetsComponent
from hydromt.model.components.tables import TablesComponent
from hydromt.model.components.vector import VectorComponent

__all__ = [
    "ConfigComponent",
    "DatasetsComponent",
    "GeomsComponent",
    "GridComponent",
    "MeshComponent",
    "ModelComponent",
    "SpatialDatasetsComponent",
    "SpatialModelComponent",
    "TablesComponent",
    "VectorComponent",
]

# define hydromt component entry points; abstract classes are not included
# see also hydromt.component group in pyproject.toml
__hydromt_eps__ = [
    "ConfigComponent",
    "DatasetsComponent",
    "GeomsComponent",
    "GridComponent",
    "MeshComponent",
    "SpatialDatasetsComponent",
    "TablesComponent",
    "VectorComponent",
]



GDAL_DRIVER_CODE_MAP = {
    "asc": "AAIGrid",
    "blx": "BLX",
    "bmp": "BMP",
    "bt": "BT",
    "dat": "ZMap",
    "dem": "USGSDEM",
    "gen": "ADRG",
    "gif": "GIF",
    "gpkg": "GPKG",
    "grd": "NWT_GRD",
    "gsb": "NTv2",
    "gtx": "GTX",
    "hdr": "MFF",
    "hf2": "HF2",
    "hgt": "SRTMHGT",
    "img": "HFA",
    "jpg": "JPEG",
    "kro": "KRO",
    "lcp": "LCP",
    "mbtiles": "MBTiles",
    "mpr/mpl": "ILWIS",
    "ntf": "NITF",
    "pix": "PCIDSK",
    "png": "PNG",
    "pnm": "PNM",
    "rda": "R",
    "rgb": "SGI",
    "rst": "RST",
    "rsw": "RMF",
    "sdat": "SAGA",
    "sqlite": "Rasterlite",
    "ter": "Terragen",
    "tif": "GTiff",
    "vrt": "VRT",
    "xpm": "XPM",
    "xyz": "XYZ",
}
GDAL_EXT_CODE_MAP = {v: k for k, v in GDAL_DRIVER_CODE_MAP.items()}



"""Example grid component module."""

import logging
from pathlib import Path
from typing import Any

import xarray as xr

from hydromt.model import Model
from hydromt.model.components import GridComponent
from hydromt.model.processes.grid import (
    create_grid_from_region,
    grid_from_rasterdataset,
)
from hydromt.model.steps import hydromt_step

__all__ = ["ExampleGridComponent"]

logger = logging.getLogger(f"hydromt.{__name__}")


class ExampleGridComponent(GridComponent):
    """Example grid component.

    Inherits from the HydroMT-core GridComponent model-component.
    It only add two hydromt_steps for creating a grid and adding data to it.
    """

    def __init__(
        self,
        model: Model,
        *,
        filename: str = "grid.nc",
        region_filename: str = "region.geojson",
    ):
        """Initialize a ExampleGridComponent.

        Parameters
        ----------
        model : Model
            HydroMT model instance
        filename : str
            The path to use for reading and writing of component data by default.
            By default "grid.nc".
        region_filename : str
            The path to use for reading and writing of the region data by default.
            By default "region.geojson".
        """
        super().__init__(
            model,
            filename=filename,
            region_component=None,
            region_filename=region_filename,
        )

    ## Create and add_data methods
    @hydromt_step
    def create_from_region(
        self,
        region: dict[str, Any],
        *,
        res: float | int | None = None,
        crs: int | None = None,
        region_crs: int = 4326,
        rotated: bool = False,
        hydrography_path: str | None = None,
        basin_index_path: str | None = None,
        add_mask: bool = True,
        align: bool = True,
        dec_origin: int = 0,
        dec_rotation: int = 3,
    ) -> xr.DataArray:
        """HYDROMT CORE METHOD: Create a 2D regular grid or reads an existing grid.

        A 2D regular grid will be created from a geometry (geom_fn) or bbox. If an
        existing grid is given, then no new grid will be generated.

        Adds/Updates model layers (if add_mask):
        * **mask** grid mask: add grid mask to grid object

        Parameters
        ----------
        region : dict
            Dictionary describing region of interest, e.g.:
            * {'bbox': [xmin, ymin, xmax, ymax]}
            * {'geom': 'path/to/polygon_geometry'}
            * {'grid': 'path/to/grid_file'}
            * {'basin': [x, y]}

            Region must be of kind [grid, bbox, geom, basin, subbasin, interbasin].
        res: float or int, optional
            Resolution used to generate 2D grid [unit of the CRS], required if region
            is not based on 'grid'.
        crs : int, optional
            EPSG code of the grid to create.
        region_crs : int, optional
            EPSG code of the region geometry, by default None. Only applies if region is
            of kind 'bbox'or if geom crs is not defined in the file itself.
        rotated : bool
            if True, a minimum rotated rectangular grid is fitted around the region,
            by default False. Only applies if region is of kind 'bbox', 'geom'
        hydrography_fn : str, optional
            Name of data source for hydrography data. Required if region is of kind
                'basin', 'subbasin' or 'interbasin'.

            * Required variables: ['flwdir'] and any other 'snapping' variable required
                to define the region.

            * Optional variables: ['basins'] if the `region` is based on a
                (sub)(inter)basins without a 'bounds' argument.

        basin_index_path: str, optional
            Name of data source with basin (bounding box) geometries associated with
            the 'basins' layer of `hydrography_fn`. Only required if the `region` is
            based on a (sub)(inter)basins without a 'bounds' argument.
        add_mask : bool
            Add mask variable to grid object, by default True.
        align : bool
            If True (default), align target transform to resolution.
        dec_origin : int, optional
            number of decimals to round the origin coordinates, by default 0
        dec_rotation : int, optional
            number of decimals to round the rotation angle, by default 3

        Returns
        -------
        grid : xr.DataArray
            Generated grid mask.
        """
        logger.info("Preparing 2D grid.")

        # Check if this component's region is a reference to another component
        if self._region_component is not None:
            raise ValueError(
                "Region is a reference to another component. Cannot create grid."
            )

        grid = create_grid_from_region(
            region,
            data_catalog=self.data_catalog,
            res=res,
            crs=crs,
            region_crs=region_crs,
            rotated=rotated,
            hydrography_path=hydrography_path,
            basin_index_path=basin_index_path,
            add_mask=add_mask,
            align=align,
            dec_origin=dec_origin,
            dec_rotation=dec_rotation,
        )
        self.set(grid)
        return grid

    @hydromt_step
    def add_data_from_rasterdataset(
        self,
        raster_data: str | Path | xr.DataArray | xr.Dataset,
        variables: list[str] | None = None,
        fill_method: str | None = None,
        reproject_method: list[str] | str | None = "nearest",
        mask_name: str | None = "mask",
        rename: dict[str, str] | None = None,
    ) -> list[str]:
        """HYDROMT CORE METHOD: Add data variable(s) from ``raster_data`` to grid component.

        If raster is a dataset, all variables will be added unless ``variables`` list
        is specified.

        Adds model layers:

        * **raster.name** grid: data from raster_data

        Parameters
        ----------
        raster_data: str, Path, xr.DataArray, xr.Dataset
            Data catalog key, path to raster file or raster xarray data object.
            If a path to a raster file is provided it will be added
            to the data_catalog with its name based on the file basename without
            extension.
        variables: list, optional
            List of variables to add to grid from raster_data. By default all.
        fill_method : str, optional
            If specified, fills nodata values using fill_nodata method.
            Available methods are {'linear', 'nearest', 'cubic', 'rio_idw'}.
        reproject_method: list, str, optional
            See rasterio.warp.reproject for existing methods, by default 'nearest'.
            Can provide a list corresponding to ``variables``.
        mask_name: str, optional
            Name of mask in self.grid to use for masking raster_data. By default 'mask'.
            Use None to disable masking.
        rename: dict, optional
            Dictionary to rename variable names in raster_data before adding to grid
            {'name_in_raster_data': 'name_in_grid'}. By default empty.

        Returns
        -------
        list
            Names of added model map layers
        """
        rename = rename or {}
        logger.info(f"Preparing grid data from raster source {raster_data}")
        # Read raster data and select variables
        ds = self.data_catalog.get_rasterdataset(
            raster_data,
            geom=self.region,
            buffer=2,
            variables=variables,
            single_var_as_array=False,
        )
        assert ds is not None
        # Data resampling
        ds_out = grid_from_rasterdataset(
            grid_like=self._get_grid_data(),
            ds=ds,
            variables=variables,
            fill_method=fill_method,
            reproject_method=reproject_method,
            mask_name=mask_name,
            rename=rename,
        )
        # Add to grid
        self.set(ds_out)



"""Utilities for logging unused kwargs."""

import logging
from typing import Any, Dict

logger = logging.getLogger(__name__)

__all__ = ["_warn_on_unused_kwargs"]


def _warn_on_unused_kwargs(obj_name: str, name_value_dict: Dict[str, Any]):
    """Warn on unused kwargs."""
    for name, value in name_value_dict.items():
        if value is not None:
            logger.debug(
                f"object: {obj_name} does not use kwarg {name} with value {value}."
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""GIS related convenience functions."""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Union

import geopandas as gpd
import numpy as np
from pyogrio import read_info
from pyproj import CRS
from pyproj.transformer import Transformer
from shapely.geometry import box
from shapely.geometry.base import BaseGeometry

from hydromt.gis.raster_utils import cellres
from hydromt.typing import Bbox, Geom, GpdShapeGeom, Zoom

__all__ = [
    "_axes_attrs",
    "_bbox_from_file_and_filters",
    "_parse_geom_bbox_buffer",
    "_to_geographic_bbox",
    "parse_crs",
    "utm_crs",
    "zoom_to_overview_level",
]


logger = logging.getLogger(__name__)


# REPROJ
def utm_crs(bbox):
    """Return wkt string of nearest UTM projects.

    Parameters
    ----------
    bbox : array-like of floats
        (xmin, ymin, xmax, ymax) bounding box in latlon WGS84 (EPSG:4326) coordinates

    Returns
    -------
    crs: pyproj.CRS
        CRS of UTM projection
    """
    left, bottom, right, top = bbox
    x = (left + right) / 2
    y = (top + bottom) / 2
    kwargs = dict(zone=int(np.ceil((x + 180) / 6)))
    # BUGFIX hydroMT v0.3.5: south=False doesn't work only add south=True if y<0
    if y < 0:
        kwargs.update(south=True)
    # BUGFIX hydroMT v0.4.6: add datum
    epsg = CRS(proj="utm", datum="WGS84", ellps="WGS84", **kwargs).to_epsg()
    return CRS.from_epsg(epsg)


def parse_crs(crs: Any, bbox: Optional[List[float]] = None) -> CRS:
    """Parse crs string to pyproj.CRS.

    Parameters
    ----------
    crs: Any
        crs int, wkt string, or pyproj.CRS object
        if crs is 'utm' the best utm zone is calculated based on the bbox
    bbox: List[float], optional
        bounding box of the data, required for 'utm' crs

    Returns
    -------
    crs: pyproj.CRS
        coordinate reference system
    """
    if crs == "utm":
        if bbox is not None:
            crs = utm_crs(bbox)
        else:
            raise ValueError('CRS "utm" requires bbox')
    else:
        crs = CRS.from_user_input(crs)
    return crs


def _axes_attrs(crs):
    """Provide CF-compliant variable names and metadata for axes.

    Parameters
    ----------
    crs: pyproj.CRS
        coordinate reference system

    Returns
    -------
    x_dim: str - variable name of x dimension (e.g. 'x')
    y_dim: str - variable name of y dimension (e.g. 'lat')
    x_attr: dict - attributes of variable x
    y_attr: dict - attributes of variable y
    """
    # # check for type of crs
    if not isinstance(crs, CRS):
        crs = CRS.from_user_input(crs)
    if crs.is_geographic:
        x_dim, y_dim = "longitude", "latitude"
    else:
        x_dim, y_dim = "x", "y"
    cf_coords = crs.cs_to_cf()
    x_attrs = [c for c in cf_coords if c["axis"] == "X"][0]
    y_attrs = [c for c in cf_coords if c["axis"] == "Y"][0]
    return x_dim, y_dim, x_attrs, y_attrs


def _parse_geom_bbox_buffer(
    geom: Optional[Geom] = None,
    bbox: Optional[Bbox] = None,
    buffer: float = 0.0,
    crs: Optional[CRS] = None,
) -> Geom:
    """Parse geom or bbox to a (buffered) geometry.

    Arguments
    ---------
    geom : geopandas.GeoDataFrame/Series, optional
        A geometry defining the area of interest.
    bbox : array-like of floats, optional
        (xmin, ymin, xmax, ymax) bounding box of area of interest
        (in WGS84 coordinates).
    buffer : float, optional
        Buffer around the `bbox` or `geom` area of interest in meters. By default 0.
    crs: pyproj.CRS, optional
        projection of the bbox or geometry. If the geometry already has a crs, this
        argument is ignored. Defaults to EPSG:4236.

    Returns
    -------
    geom: geometry
        the actual geometry
    """
    if crs is None:
        crs = CRS("EPSG:4326")
    if geom is None and bbox is not None:
        # convert bbox to geom with crs EPGS:4326 to apply buffer later
        geom = gpd.GeoDataFrame(geometry=[box(*bbox)], crs=crs)
    elif geom is None:
        raise ValueError("No geom or bbox provided.")
    elif geom.crs is None:
        geom = geom.set_crs(crs)

    if buffer > 0:
        # make sure geom is projected > buffer in meters!
        if geom.crs.is_geographic:
            geom = geom.to_crs(3857)
        geom = geom.buffer(buffer)
    return geom


def _to_geographic_bbox(
    bbox: List[float], source_crs: Optional[CRS] = None
) -> List[float]:
    """Convert a bbox to geographic coordinates (EPSG:4326).

    Parameters
    ----------
    bbox: List[float]
        (xmin, ymin, xmax, ymax) bounding box in the source crs.
    source_crs: pyproj.CRS, optional
        coordinate reference system of the bounding box. If not provided, the
        bounding box is assumed to be in WGS84 (EPSG:4326).

    Returns
    -------
    bbox: List[float]
        (xmin, ymin, xmax, ymax) bounding box in geographic coordinates (EPSG:4326).
    """
    target_crs = CRS.from_user_input(4326)
    if source_crs is None:
        logger.warning("No CRS was set. Skipping CRS conversion")
    elif source_crs != target_crs:
        bbox = Transformer.from_crs(source_crs, target_crs).transform_bounds(*bbox)

    return bbox


def _bbox_from_file_and_filters(
    path: str,
    bbox: Optional[GpdShapeGeom] = None,
    mask: Optional[GpdShapeGeom] = None,
    crs: Optional[CRS] = None,
) -> Optional[Bbox]:
    """Create a bbox from the file metadata and filter options.

    Pyogrio does not accept a mask, and requires a bbox in the same CRS as the data.
    This function takes the possible bbox filter, mask filter and crs of the input data
    and returns a bbox in the same crs as the data based on the input filters.
    As pyogrio currently does not support filtering using a mask, the mask is converted
    to a bbox and the bbox is returned so that the data has some geospatial filtering.

    Parameters
    ----------
    path: IOBase,
        path to the opened file.
    bbox: GeoDataFrame | GeoSeries | BaseGeometry
        bounding box to filter the data while reading.
    mask: GeoDataFrame | GeoSeries | BaseGeometry
        mask to filter the data while reading.
    crs: pyproj.CRS
        coordinate reference system of the bounding box or geometry. If already set,
        this argument is ignored.
    """
    if bbox is not None and mask is not None:
        raise ValueError(
            "Both 'bbox' and 'mask' are provided. Please provide only one."
        )
    if bbox is None and mask is None:
        return None
    if source_crs_str := read_info(path).get("crs"):
        source_crs = CRS(source_crs_str)
    elif crs:
        source_crs = crs
    else:  # assume WGS84
        source_crs = CRS("EPSG:4326")

    if mask is not None:
        bbox = mask

    # convert bbox to geom with input crs (assume WGS84 if not provided)
    crs = crs if crs is not None else CRS.from_user_input(4326)
    if issubclass(type(bbox), BaseGeometry):
        bbox = gpd.GeoSeries(bbox, crs=crs)
    bbox = bbox if bbox.crs is not None else bbox.set_crs(crs)
    return tuple(bbox.to_crs(source_crs).total_bounds)


def zoom_to_overview_level(
    zoom: Zoom,
    mask: Optional[Geom] = None,
    zls_dict: Optional[Dict[int, float]] = None,
    source_crs: Union[CRS, int, None] = None,
) -> Optional[int]:
    """Return overview level of data corresponding to zoom level.

    Parameters
    ----------
    zoom: Zoom
        Overview level or tuple with resolution and unit
    mask: gpd.GeoSeries, optional
        Geometry to determine res if zoom_level or source in degree
    zls_dict: Dict[int, float], optional
        Dictionary translating overview level to resolution.
    source_crs: pyproj.CRS, optional
        Source crs to determine res if zoom_level tuple is provided
        with different unit than source_crs

    Returns
    -------
    overview_level: int or None
        Overview level of the data to be used. If no overview levels are defined,
        None is returned.
    """
    # check zoom level
    if zls_dict is None or len(zls_dict) == 0:
        return None
    elif isinstance(zoom, int):
        overview_level = zoom
        if overview_level not in zls_dict:
            raise ValueError(
                f"Overview level {overview_level} not defined.Select from {zls_dict}."
            )
        dst_res = zls_dict[overview_level]
    elif (
        isinstance(zoom, tuple)
        and isinstance(zoom[0], (int, float))
        and isinstance(zoom[1], str)
        and len(zoom) == 2
        and source_crs is not None
    ):
        src_res, src_res_unit = zoom
        # convert res if different unit than crs
        source_crs = CRS.from_user_input(source_crs)
        dst_crs_unit = source_crs.axis_info[0].unit_name
        dst_res = src_res
        if dst_crs_unit != src_res_unit:
            known_units = ["degree", "metre", "US survey foot", "meter", "foot"]
            if src_res_unit not in known_units:
                raise TypeError(
                    f"zoom_level unit {src_res_unit} not understood;"
                    f" should be one of {known_units}"
                )
            if dst_crs_unit not in known_units:
                raise NotImplementedError(
                    f"no conversion available for {src_res_unit} to {dst_crs_unit}"
                )
            conversions = {
                "foot": 0.3048,
                "metre": 1,  # official pyproj units
                "US survey foot": 0.3048,  # official pyproj units
            }  # to meter
            if src_res_unit == "degree" or dst_crs_unit == "degree":
                lat = 0
                if mask is not None:
                    lat = mask.to_crs(4326).centroid.y.item()
                conversions["degree"] = cellres(lat=lat)[1]
            fsrc = conversions.get(src_res_unit, 1)
            fdst = conversions.get(dst_crs_unit, 1)
            dst_res = src_res * fsrc / fdst
        # find nearest zoom level
        res = list(zls_dict.values())[0] / 2
        zls = list(zls_dict.keys())
        smaller = [x < (dst_res + res * 0.01) for x in zls_dict.values()]
        overview_level = (
            zls[-1] if all(smaller) else zls[max(smaller.index(False) - 1, 0)]
        )
    elif source_crs is None:
        raise ValueError("No CRS defined, hence no zoom level can be determined.")
    else:
        raise TypeError(f"zoom_level not understood: {zoom}")
    logger.debug(f"Using overview level {overview_level} ({dst_res:.2f})")
    return overview_level



"""Implementations for all of the necessary IO writing for HydroMT."""

import hashlib
import logging
import uuid
from contextlib import nullcontext
from pathlib import Path
from typing import Any, cast

import geopandas as gpd
import numpy as np
import xarray as xr
from dask.diagnostics import ProgressBar
from tomli_w import dump as dump_toml
from yaml import dump as dump_yaml

from hydromt.gis import _normalize
from hydromt.typing.deferred_file_close import DeferredFileClose

logger = logging.getLogger(__name__)

__all__ = [
    "write_nc",
    "write_region",
    "write_toml",
    "write_xy",
    "write_yaml",
]


def write_yaml(path: str | Path, data: dict[str, Any], **kwargs):
    """Write a dictionary to a yaml formatted file."""
    with open(path, "w") as f:
        dump_yaml(data, f, **kwargs)


def write_toml(path: str | Path, data: dict[str, Any]):
    """Write a dictionary to a toml formatted file."""
    with open(path, "wb") as f:
        dump_toml(data, f)


def write_xy(path: str | Path, gdf, fmt="%.4f"):
    """Write geopandas.GeoDataFrame with Point geometries to point xy files.

    Parameters
    ----------
    path: str or Path
        Path to the output file.
    gdf: geopandas.GeoDataFrame
        GeoDataFrame to write to point file.
    fmt: fmt
        String formatting. By default "%.4f".
    """
    if not np.all(gdf.geometry.type == "Point"):
        raise ValueError("gdf should contain only Point geometries.")
    xy = np.stack((gdf.geometry.x.values, gdf.geometry.y.values)).T
    with open(path, "w") as f:
        np.savetxt(f, xy, fmt=fmt)


def write_nc(
    ds: xr.DataArray | xr.Dataset,
    file_path: Path,
    *,
    compress: bool = False,
    gdal_compliant: bool = False,
    rename_dims: bool = False,
    force_sn: bool = False,
    force_overwrite: bool = False,
    progressbar: bool = False,
    to_netcdf_kwargs: dict[str, Any] | None = None,
) -> DeferredFileClose | None:
    """Write xarray.Dataset and/or xarray.DataArray to netcdf file.

    Possibility to update the xarray objects attributes to get GDAL compliant NetCDF
    files, using :py:meth:`~hydromt.raster.gdal_compliant`.
    The function will first try to directly write to file. In case of
    PermissionError, it will first write a temporary file and move the file over.

    key-word arguments are passed to :py:meth:`xarray.Dataset.to_netcdf`

    Parameters
    ----------
    ds : xr.DataArray | xr.Dataset
        Dataset to be written to the drive
    file_path : Path
        Full path to the outgoing file
    compress : bool, optional
        Whether or not to compress the data, by default False
    gdal_compliant : bool, optional
        If True, convert xarray.Dataset and/or xarray.DataArray to gdal compliant
        format using :py:meth:`~hydromt.raster.gdal_compliant`, by default False
    rename_dims : bool, optional
        If True, rename x_dim and y_dim to standard names depending on the CRS
        (x/y for projected and lat/lon for geographic). Only used if
        ``gdal_compliant`` is set to True. By default False
    force_overwrite : bool, optional
        If True, forces the file to be overwritten if it already exists. By default False
    force_sn : bool, optional
        If True, forces the dataset to have South -> North orientation. Only used
        if ``gdal_compliant`` is set to True. By default False
    progressbar : bool, optional
        If True, the netcdf will be computed and written with a visible progressbar.
        This is only useful when the dataset to be written is lazily read and modified
        using dask. By default False
    to_netcdf_kwargs : dict
        Additional keyword arguments that are passed to the `to_netcdf` function.
    """
    # Check the typing
    if not isinstance(ds, (xr.Dataset, xr.DataArray)) or len(ds) == 0:
        logger.error(f"Dataset object of type {type(ds).__name__} not recognized")
        return None
    if isinstance(ds, xr.DataArray):
        if ds.name is None:
            ds.name = file_path.stem
        ds = ds.to_dataset()
    # Check whether the file already exists
    if file_path.exists() and not force_overwrite:
        raise IOError(f"File {file_path.as_posix()} already exists")
    file_path.parent.mkdir(parents=True, exist_ok=True)

    # Focus on the encoding and set these for all dims, coords and data vars
    to_netcdf_kwargs = to_netcdf_kwargs or {}
    to_netcdf_kwargs = to_netcdf_kwargs.copy()
    encoding = to_netcdf_kwargs.pop("encoding", {})

    for var in set(ds.coords) | set(ds.data_vars):
        if var not in encoding:
            encoding[var] = {}

    # Remove the nodata val specifier for the dimensions, CF compliant that is
    for dim in ds.dims:
        ds[dim].attrs.pop("_FillValue", None)
        if dim in encoding:
            encoding[dim].update({"_FillValue": None})

    # Set compression if True
    if compress:
        for var in ds.data_vars:
            encoding[var].update({"zlib": True})

    # Make gdal compliant if True, only in case of a spatial dataset
    if gdal_compliant:
        y_old, x_old = ds.raster.dims
        ds = ds.raster.gdal_compliant(rename_dims=rename_dims, force_sn=force_sn)
        y_dim, x_dim = ds.raster.dims
        encoding[y_dim] = encoding.pop(y_old)
        encoding[x_dim] = encoding.pop(x_old)

    to_netcdf_kwargs["encoding"] = encoding

    # Try to write the file
    try:
        _nc_progress(
            ds,
            file_path=file_path,
            progressbar=progressbar,
            to_netcdf_kwargs=to_netcdf_kwargs,
        )
    except OSError:
        logger.debug(f"Could not write to file {file_path.as_posix()}, deferring write")
        unique_str = f"{file_path}_{uuid.uuid4()}"
        hash_str = hashlib.sha256(unique_str.encode()).hexdigest()[:8]
        temp_filepath = file_path.with_stem(f"{file_path.stem}_{hash_str}")
        _nc_progress(
            ds,
            file_path=temp_filepath,
            progressbar=progressbar,
            to_netcdf_kwargs=to_netcdf_kwargs,
        )
        return DeferredFileClose(original_path=file_path, temp_path=temp_filepath)

    return None


def write_region(
    region: gpd.GeoDataFrame,
    file_path: Path,
    *,
    to_wgs84=False,
    to_file_kwargs: dict[str, Any] | None = None,
):
    """Write the model region to a file.

    Parameters
    ----------
    file_path : Path
        The filename to write the region to.
    to_wgs84 : bool, optional
        If True, the region is reprojected to WGS84 before writing. default is False.
    to_file_kwargs: dict, optional
        Additional keyword arguments passed to the `geopandas.GeoDataFrame.to_file` function.
    """
    file_path.parent.mkdir(parents=True, exist_ok=True)
    to_file_kwargs = to_file_kwargs or {}

    gdf = cast(gpd.GeoDataFrame, region.copy())

    if to_wgs84 and (
        to_file_kwargs.get("driver") == "GeoJSON"
        or file_path.suffix.lower() == ".geojson"
    ):
        gdf = gdf.to_crs(4326)

    gdf.to_file(file_path, **to_file_kwargs)


def _nc_progress(
    ds: xr.Dataset | xr.DataArray,
    *,
    file_path: Path,
    progressbar: bool,
    to_netcdf_kwargs: dict[str, Any],
):
    if "compute" in to_netcdf_kwargs:
        raise ValueError(
            "'compute' argument is ignored in ds.to_netcdf function. Did you mean to use 'progressbar'?"
        )
    ds = _normalize.normalize_xarray_dtypes(ds)
    obj = ds.to_netcdf(file_path, compute=False, **to_netcdf_kwargs)
    with ProgressBar() if progressbar else nullcontext():
        obj.compute()
    # Dereference
    obj = None



import re
from string import Formatter

_PLACEHOLDERS = frozenset({"year", "month", "variable", "name", "overview_level"})
_SEGMENT_PATTERN = r"[^/\\]+"


def _expand_uri_placeholders(
    uri: str,
    *,
    placeholders: list[str] | None = None,
    time_range: tuple[str, str] | None = None,
    variables: list[str] | None = None,
) -> tuple[str, list[str], re.Pattern[str]]:
    """Expand known placeholders in the URI.

    This function takes a URI with placeholders and expands it into a regex pattern.
    It also returns a list of keys corresponding to the placeholders that were captured in the regex.

    Parameters
    ----------
    uri : str
        The URI containing placeholders to be expanded. Should not contain any regex special characters, as they will be escaped.
    placeholders : list[str] | None, optional
        A list of placeholder names to capture in the regex.
        If None, no placeholders will be captured, but they will still be expanded in the URI.
        Default is None.
    time_range : tuple[str, str] | None, optional
        A tuple containing the start and end of a time range (e.g., ("2020", "2021")).
        If provided, the function will not expand "year" and "month" placeholders to wildcards.
        Default is None.
    variables : list[str] | None, optional
        A list of variable names to capture in the regex.
        If provided, the function will not expand the "variable" placeholder to a wildcard.
        Default is None.

    Returns
    -------
    tuple[str, list[str], Pattern[str]]
        A tuple containing the expanded URI, a list of keys corresponding to the captured placeholders, and a compiled regex pattern for matching the expanded URI.
    """
    if placeholders is None:
        placeholders = []
    keys: list[str] = []
    pattern: str = ""
    uri_expanded = ""

    if "{" in uri:
        for literal_text, key, fmt, _ in Formatter().parse(uri):
            safe_literal = (
                re.escape(literal_text).replace(r"\*", ".*").replace(r"\?", ".")
            )
            pattern += safe_literal
            uri_expanded += literal_text
            if key is None:
                continue

            if key in placeholders:
                pattern += (
                    f"({_SEGMENT_PATTERN})"  # capture only requested placeholders
                )
                keys.append(key)
            else:
                pattern += f"(?:{_SEGMENT_PATTERN})"  # match segment, but don't capture

            key_str = "{" + f"{key}:{fmt}" + "}" if fmt else "{" + key + "}"
            # Determine if this key should become a wildcard
            if key in ["year", "month"] and time_range is None:
                uri_expanded += "*"
            elif key == "variable" and variables is None:
                uri_expanded += "*"
            elif key == "name":
                uri_expanded += "*"
            else:
                uri_expanded += key_str
        uri = uri_expanded

    # Anchor the regex to make sure the entire path matches, not just a substring
    pattern = f"^{pattern}$"

    # windows paths creating invalid escape sequences
    try:
        regex = re.compile(pattern)
    except re.error:
        # try it as raw path if regular string fails
        regex = re.compile(pattern.encode("unicode_escape").decode())

    return (uri, keys, regex)



"""Coordinate Reference System type based on pyproj.CRS with pydantic validation and serialization."""

from typing import Any, List, Optional, Union

from pydantic import PlainSerializer, PlainValidator
from pyproj import CRS as ProjCRS
from pyproj.exceptions import ProjError
from typing_extensions import Annotated


def _serialize_crs(crs: ProjCRS) -> Union[int, List[str], str]:
    # Try epsg authority first
    epsg: Optional[int] = crs.to_epsg()
    if not epsg:
        # Then try any authority
        auth: Optional[List[str, str]] = list(
            crs.to_authority()
        )  # cast to list to be serializable to yaml
        if not auth:
            # reserve wkt for last
            return crs.to_wkt()
        return auth
    return epsg


def _validate_crs(crs: Any) -> ProjCRS:
    try:
        return ProjCRS.from_user_input(crs)
    except ProjError:
        return ProjCRS.from_authority("ESRI", crs)  # fallback on ESRI


CRS = Annotated[ProjCRS, PlainValidator(_validate_crs), PlainSerializer(_serialize_crs)]
